import py, weakref
from rpython.jit.backend import model
from rpython.jit.backend.llgraph import support
from rpython.jit.backend.llsupport import symbolic
from rpython.jit.backend.llsupport.vector_ext import VectorExt
from rpython.jit.metainterp.history import BackendDescr
from rpython.jit.metainterp.history import Const, getkind
from rpython.jit.metainterp.history import INT, REF, FLOAT, VOID
from rpython.jit.metainterp.resoperation import rop
from rpython.jit.metainterp.optimizeopt import intbounds
from rpython.jit.metainterp.optimize import SpeculativeError
from rpython.jit.metainterp.support import ptr2int, int_signext
from rpython.jit.codewriter import longlong, heaptracker
from rpython.jit.codewriter.effectinfo import EffectInfo

from rpython.rtyper.llinterp import LLInterpreter, LLException
from rpython.rtyper.lltypesystem import lltype, llmemory, rffi, rstr
from rpython.rtyper.lltypesystem.lloperation import llop
from rpython.rtyper.annlowlevel import hlstr, hlunicode
from rpython.rtyper import rclass

from rpython.rlib.clibffi import FFI_DEFAULT_ABI
from rpython.rlib.rarithmetic import ovfcheck, r_uint, r_ulonglong, intmask
from rpython.rlib.objectmodel import Symbolic, compute_hash

class LLAsmInfo(object):
    def __init__(self, lltrace):
        self.ops_offset = None
        self.lltrace = lltrace

class LLTrace(object):
    has_been_freed = False
    invalid = False

    def __init__(self, inputargs, operations):
        # We need to clone the list of operations because the
        # front-end will mutate them under our feet again.  We also
        # need to make sure things get freed.
        _cache={}

        def mapping(box):
            if isinstance(box, Const) or box is None:
                return box
            try:
                newbox = _cache[box]
            except KeyError:
                newbox = _cache[box] = box.__class__()
            if hasattr(box, 'accum') and box.accum:
                newbox.accum = box.accum
            return newbox
        #
        self.inputargs = map(mapping, inputargs)
        self.operations = []
        for op in operations:
            opnum = op.getopnum()
            if opnum == rop.GUARD_VALUE:
                # we don't care about the value 13 here, because we gonna
                # fish it from the extra slot on frame anyway
                op.getdescr().make_a_counter_per_value(op, 13)
            if op.getdescr() is not None:
                if op.is_guard() or op.getopnum() == rop.FINISH:
                    newdescr = op.getdescr()
                else:
                    newdescr = WeakrefDescr(op.getdescr())
            else:
                newdescr = None
            newop = op.copy_and_change(op.getopnum(),
                                       map(mapping, op.getarglist()),
                                       newdescr)
            _cache[op] = newop
            if op.getfailargs() is not None:
                newop.setfailargs(map(mapping, op.getfailargs()))
            self.operations.append(newop)

class WeakrefDescr(BackendDescr):
    def __init__(self, realdescr):
        self.realdescrref = weakref.ref(realdescr)
        self.final_descr = getattr(realdescr, 'final_descr', False)

class ExecutionFinished(Exception):
    def __init__(self, deadframe):
        self.deadframe = deadframe

class Jump(Exception):
    def __init__(self, jump_target, args):
        self.jump_target = jump_target
        self.args = args

class CallDescr(BackendDescr):
    def __init__(self, RESULT, ARGS, extrainfo, ABI=FFI_DEFAULT_ABI):
        self.RESULT = RESULT
        self.ARGS = ARGS
        self.ABI = ABI
        self.extrainfo = extrainfo

    def __repr__(self):
        return 'CallDescr(%r, %r, %r)' % (self.RESULT, self.ARGS,
                                          self.extrainfo)

    def get_extra_info(self):
        return self.extrainfo

    def get_arg_types(self):
        return ''.join([getkind(ARG)[0] for ARG in self.ARGS])

    def get_result_type(self):
        return getkind(self.RESULT)[0]

    get_normalized_result_type = get_result_type

class TypeIDSymbolic(Symbolic):
    def __init__(self, STRUCT_OR_ARRAY):
        self.STRUCT_OR_ARRAY = STRUCT_OR_ARRAY

    def __eq__(self, other):
        return self.STRUCT_OR_ARRAY is other.STRUCT_OR_ARRAY

    def __ne__(self, other):
        return not self == other

class SizeDescr(BackendDescr):
    def __init__(self, S, vtable, runner):
        assert not isinstance(vtable, bool)
        self.S = S
        self._vtable = vtable
        self._is_object = bool(vtable)
        self._runner = runner

    def get_all_fielddescrs(self):
        return self.all_fielddescrs

    def is_object(self):
        return self._is_object

    def get_vtable(self):
        assert self._vtable is not None
        if self._vtable is Ellipsis:
            self._vtable = heaptracker.get_vtable_for_gcstruct(self._runner,
                                                               self.S)
        return ptr2int(self._vtable)

    def is_immutable(self):
        return heaptracker.is_immutable_struct(self.S)

    def get_type_id(self):
        assert isinstance(self.S, lltype.GcStruct)
        return TypeIDSymbolic(self.S)     # integer-like symbolic

    def __repr__(self):
        return 'SizeDescr(%r)' % (self.S,)

class FieldDescr(BackendDescr):
    def __init__(self, S, fieldname):
        self.S = S
        self.fieldname = fieldname
        self.FIELD = getattr(S, fieldname)
        self.index = heaptracker.get_fielddescr_index_in(S, fieldname)
        self._is_pure = S._immutable_field(fieldname) != False

    def is_always_pure(self):
        return self._is_pure

    def get_parent_descr(self):
        return self.parent_descr

    def get_vinfo(self):
        return self.vinfo

    def get_index(self):
        return self.index

    def __repr__(self):
        return 'FieldDescr(%r, %r)' % (self.S, self.fieldname)

    def sort_key(self):
        return self.fieldname

    def is_pointer_field(self):
        return getkind(self.FIELD) == 'ref'

    def is_float_field(self):
        return getkind(self.FIELD) == 'float'

    def is_field_signed(self):
        return _is_signed_kind(self.FIELD)

    def is_integer_bounded(self):
        return getkind(self.FIELD) == 'int' \
            and rffi.sizeof(self.FIELD) < symbolic.WORD

    def get_integer_min(self):
        if getkind(self.FIELD) != 'int':
            assert False

        return intbounds.get_integer_min(
            not _is_signed_kind(self.FIELD), rffi.sizeof(self.FIELD))

    def get_integer_max(self):
        if getkind(self.FIELD) != 'int':
            assert False

        return intbounds.get_integer_max(
            not _is_signed_kind(self.FIELD), rffi.sizeof(self.FIELD))


def _is_signed_kind(TYPE):
    return (TYPE is not lltype.Bool and isinstance(TYPE, lltype.Number) and
            rffi.cast(TYPE, -1) == -1)

class ArrayDescr(BackendDescr):
    all_interiorfielddescrs = None

    def __init__(self, A, runner):
        self.A = self.OUTERA = A
        self._is_pure = A._immutable_field(None)
        self.concrete_type = '\x00'
        if isinstance(A, lltype.Struct):
            self.A = A._flds[A._arrayfld]

    def is_array_of_primitives(self):
        kind = getkind(self.A.OF)
        return kind == 'float' or \
               kind == 'int'

    def is_always_pure(self):
        return self._is_pure

    def get_all_fielddescrs(self):
        return self.all_interiorfielddescrs

    def __repr__(self):
        return 'ArrayDescr(%r)' % (self.OUTERA,)

    def is_array_of_pointers(self):
        return getkind(self.A.OF) == 'ref'

    def is_array_of_floats(self):
        return getkind(self.A.OF) == 'float'

    def is_item_signed(self):
        return _is_signed_kind(self.A.OF)

    def is_array_of_structs(self):
        return isinstance(self.A.OF, lltype.Struct)

    def is_item_integer_bounded(self):
        return getkind(self.A.OF) == 'int' \
            and rffi.sizeof(self.A.OF) < symbolic.WORD

    def get_item_size_in_bytes(self):
        return rffi.sizeof(self.A.OF)

    def get_item_integer_min(self):
        if getkind(self.A.OF) != 'int':
            assert False

        return intbounds.get_integer_min(
            not _is_signed_kind(self.A.OF), rffi.sizeof(self.A.OF))

    def get_item_integer_max(self):
        if getkind(self.A.OF) != 'int':
            assert False

        return intbounds.get_integer_max(
            not _is_signed_kind(self.A.OF), rffi.sizeof(self.A.OF))

    def get_type_id(self):
        assert isinstance(self.A, lltype.GcArray)
        return TypeIDSymbolic(self.A)     # integer-like symbolic


class InteriorFieldDescr(BackendDescr):
    def __init__(self, A, fieldname, runner):
        self.A = A
        self.fieldname = fieldname
        self.FIELD = getattr(A.OF, fieldname)
        self.arraydescr = runner.arraydescrof(A)
        self.fielddescr = runner.fielddescrof(A.OF, fieldname)

    def get_index(self):
        return self.fielddescr.get_index()

    def get_arraydescr(self):
        return self.arraydescr

    def get_field_descr(self):
        return self.fielddescr

    def __repr__(self):
        return 'InteriorFieldDescr(%r, %r)' % (self.A, self.fieldname)

    def sort_key(self):
        return self.fieldname

    def is_pointer_field(self):
        return getkind(self.FIELD) == 'ref'

    def is_float_field(self):
        return getkind(self.FIELD) == 'float'

    def is_integer_bounded(self):
        return getkind(self.FIELD) == 'int' \
            and rffi.sizeof(self.FIELD) < symbolic.WORD

    def get_integer_min(self):
        if getkind(self.FIELD) != 'int':
            assert False

        return intbounds.get_integer_min(
            not _is_signed_kind(self.FIELD), rffi.sizeof(self.FIELD))

    def get_integer_max(self):
        if getkind(self.FIELD) != 'int':
            assert False

        return intbounds.get_integer_max(
            not _is_signed_kind(self.FIELD), rffi.sizeof(self.FIELD))

_example_res = {'v': None,
                'r': lltype.nullptr(llmemory.GCREF.TO),
                'i': 0,
                'f': 0.0}


class LLGraphCPU(model.AbstractCPU):
    supports_floats = True
    supports_longlong = r_uint is not r_ulonglong
    supports_singlefloats = True
    supports_guard_gc_type = True
    translate_support_code = False
    is_llgraph = True
    vector_ext = VectorExt()
    vector_ext.enable(16, accum=True)
    vector_ext.setup_once = lambda asm: asm
    load_supported_factors = (1,2,4,8)
    assembler = None

    def __init__(self, rtyper, stats=None, *ignored_args, **kwds):
        model.AbstractCPU.__init__(self)
        self.rtyper = rtyper
        self.llinterp = LLInterpreter(rtyper)
        self.descrs = {}
        class MiniStats:
            pass
        self.stats = stats or MiniStats()
        self.vinfo_for_tests = kwds.get('vinfo_for_tests', None)

    def stitch_bridge(self, faildescr, target):
        faildescr._llgraph_bridge = target[0].lltrace

    def compile_loop(self, inputargs, operations, looptoken, jd_id=0,
                     unique_id=0, log=True, name='', logger=None):
        clt = model.CompiledLoopToken(self, looptoken.number)
        looptoken.compiled_loop_token = clt
        lltrace = LLTrace(inputargs, operations)
        clt._llgraph_loop = lltrace
        clt._llgraph_alltraces = [lltrace]
        self._record_labels(lltrace)

    def compile_bridge(self, faildescr, inputargs, operations,
                       original_loop_token, log=True, logger=None):
        clt = original_loop_token.compiled_loop_token
        clt.compiling_a_bridge()
        lltrace = LLTrace(inputargs, operations)
        faildescr._llgraph_bridge = lltrace
        clt._llgraph_alltraces.append(lltrace)
        self._record_labels(lltrace)
        return LLAsmInfo(lltrace)

    def _record_labels(self, lltrace):
        for i, op in enumerate(lltrace.operations):
            if op.getopnum() == rop.LABEL:
                _getdescr(op)._llgraph_target = (lltrace, i)

    def invalidate_loop(self, looptoken):
        for trace in looptoken.compiled_loop_token._llgraph_alltraces:
            trace.invalid = True

    def redirect_call_assembler(self, oldlooptoken, newlooptoken):
        oldc = oldlooptoken.compiled_loop_token
        newc = newlooptoken.compiled_loop_token
        oldtrace = oldc._llgraph_loop
        newtrace = newc._llgraph_loop
        OLD = [box.type for box in oldtrace.inputargs]
        NEW = [box.type for box in newtrace.inputargs]
        assert OLD == NEW
        assert not hasattr(oldc, '_llgraph_redirected')
        oldc._llgraph_redirected = newc
        oldc._llgraph_alltraces = newc._llgraph_alltraces

    def free_loop_and_bridges(self, compiled_loop_token):
        for c in compiled_loop_token._llgraph_alltraces:
            c.has_been_freed = True
        compiled_loop_token._llgraph_alltraces = []
        compiled_loop_token._llgraph_loop = None
        model.AbstractCPU.free_loop_and_bridges(self, compiled_loop_token)

    def make_execute_token(self, *argtypes):
        return self._execute_token

    def _execute_token(self, loop_token, *args):
        loopc = loop_token.compiled_loop_token
        while hasattr(loopc, '_llgraph_redirected'):
            loopc = loopc._llgraph_redirected
        lltrace = loopc._llgraph_loop
        frame = LLFrame(self, lltrace.inputargs, args)
        try:
            frame.execute(lltrace)
            assert False
        except ExecutionFinished as e:
            return e.deadframe

    def get_value_direct(self, deadframe, tp, index):
        v = deadframe._extra_value
        if tp == 'i':
            assert lltype.typeOf(v) == lltype.Signed
        elif tp == 'r':
            assert lltype.typeOf(v) == llmemory.GCREF
        elif tp == 'f':
            assert lltype.typeOf(v) == longlong.FLOATSTORAGE
        else:
            assert False
        return v

    def get_int_value(self, deadframe, index):
        v = deadframe._values[index]
        assert lltype.typeOf(v) == lltype.Signed
        return v

    def get_ref_value(self, deadframe, index):
        v = deadframe._values[index]
        assert lltype.typeOf(v) == llmemory.GCREF
        return v

    def get_float_value(self, deadframe, index):
        v = deadframe._values[index]
        assert lltype.typeOf(v) == longlong.FLOATSTORAGE
        return v

    def get_latest_descr(self, deadframe):
        return deadframe._latest_descr

    def grab_exc_value(self, deadframe):
        if deadframe._last_exception is not None:
            result = deadframe._last_exception.args[1]
            gcref = lltype.cast_opaque_ptr(llmemory.GCREF, result)
        else:
            gcref = lltype.nullptr(llmemory.GCREF.TO)
        return gcref

    def force(self, force_token):
        frame = force_token
        assert isinstance(frame, LLFrame)
        assert frame.forced_deadframe is None
        values = []
        for box in frame.force_guard_op.getfailargs():
            if box is not None:
                if box is not frame.current_op:
                    value = frame.env[box]
                else:
                    value = 0 # box.getvalue()    # 0 or 0.0 or NULL
            else:
                value = None
            values.append(value)
        frame.forced_deadframe = LLDeadFrame(
            _getdescr(frame.force_guard_op), values)
        return frame.forced_deadframe

    def set_savedata_ref(self, deadframe, data):
        deadframe._saved_data = data

    def get_savedata_ref(self, deadframe):
        assert deadframe._saved_data is not None
        return deadframe._saved_data

    # ------------------------------------------------------------

    def setup_descrs(self):
        all_descrs = []
        for k, v in self.descrs.iteritems():
            v.descr_index = len(all_descrs)
            all_descrs.append(v)
        return all_descrs

    def fetch_all_descrs(self):
        return self.descrs.values()

    def calldescrof(self, FUNC, ARGS, RESULT, effect_info):
        key = ('call', getkind(RESULT),
               tuple([getkind(A) for A in ARGS]),
               effect_info)
        try:
            return self.descrs[key]
        except KeyError:
            descr = CallDescr(RESULT, ARGS, effect_info)
            self.descrs[key] = descr
            return descr

    def sizeof(self, S, vtable=lltype.nullptr(rclass.OBJECT_VTABLE)):
        key = ('size', S)
        try:
            descr = self.descrs[key]
        except KeyError:
            descr = SizeDescr(S, vtable, self)
            self.descrs[key] = descr
            descr.all_fielddescrs = heaptracker.all_fielddescrs(self, S,
                    get_field_descr=LLGraphCPU.fielddescrof)
        if descr._is_object and vtable is not Ellipsis:
            assert vtable
            heaptracker.testing_gcstruct2vtable.setdefault(S, vtable)
        return descr

    def fielddescrof(self, S, fieldname):
        key = ('field', S, fieldname)
        try:
            return self.descrs[key]
        except KeyError:
            descr = FieldDescr(S, fieldname)
            self.descrs[key] = descr
            if (isinstance(S, lltype.GcStruct) and
                    heaptracker.has_gcstruct_a_vtable(S)):
                vtable = Ellipsis
            else:
                vtable = None
            descr.parent_descr = self.sizeof(S, vtable)
            if self.vinfo_for_tests is not None:
                descr.vinfo = self.vinfo_for_tests
            return descr

    def arraydescrof(self, A):
        key = ('array', A)
        try:
            return self.descrs[key]
        except KeyError:
            descr = ArrayDescr(A, self)
            self.descrs[key] = descr
            if isinstance(A, lltype.Array) and isinstance(A.OF, lltype.Struct):
                descrs = heaptracker.all_interiorfielddescrs(self,
                        A, get_field_descr=LLGraphCPU.interiorfielddescrof)
                descr.all_interiorfielddescrs = descrs
            return descr

    def interiorfielddescrof(self, A, fieldname):
        key = ('interiorfield', A, fieldname)
        try:
            return self.descrs[key]
        except KeyError:
            descr = InteriorFieldDescr(A, fieldname, self)
            self.descrs[key] = descr
            return descr

    def _calldescr_dynamic_for_tests(self, atypes, rtype,
                                     abiname='FFI_DEFAULT_ABI'):
        # XXX WTF is that and why it breaks all abstractions?
        from rpython.jit.backend.llsupport import ffisupport
        return ffisupport.calldescr_dynamic_for_tests(self, atypes, rtype,
                                                      abiname)

    def calldescrof_dynamic(self, cif_description, extrainfo):
        # XXX WTF, this is happy nonsense
        from rpython.jit.backend.llsupport.ffisupport import get_ffi_type_kind
        from rpython.jit.backend.llsupport.ffisupport import UnsupportedKind
        ARGS = []
        try:
            for itp in range(cif_description.nargs):
                arg = cif_description.atypes[itp]
                kind = get_ffi_type_kind(self, arg)
                if kind != VOID:
                    ARGS.append(support.kind2TYPE[kind[0]])
            RESULT = support.kind2TYPE[get_ffi_type_kind(self, cif_description.rtype)[0]]
        except UnsupportedKind:
            return None
        key = ('call_dynamic', RESULT, tuple(ARGS),
               extrainfo, cif_description.abi)
        try:
            return self.descrs[key]
        except KeyError:
            descr = CallDescr(RESULT, ARGS, extrainfo, ABI=cif_description.abi)
            self.descrs[key] = descr
            return descr

    def check_is_object(self, gcptr):
        """Check if the given, non-null gcptr refers to an rclass.OBJECT
        or not at all (an unrelated GcStruct or a GcArray).  Only usable
        in the llgraph backend, or after translation of a real backend."""
        ptr = lltype.normalizeptr(gcptr._obj.container._as_ptr())
        T = lltype.typeOf(ptr).TO
        return heaptracker.has_gcstruct_a_vtable(T) or T is rclass.OBJECT

    def get_actual_typeid(self, gcptr):
        """Fetch the actual typeid of the given gcptr, as an integer.
        Only usable in the llgraph backend, or after translation of a
        real backend.  (Here in the llgraph backend, returns a
        TypeIDSymbolic instead of a real integer.)"""
        ptr = lltype.normalizeptr(gcptr._obj.container._as_ptr())
        return TypeIDSymbolic(lltype.typeOf(ptr).TO)

    # ------------------------------------------------------------

    def maybe_on_top_of_llinterp(self, func, args, RESULT):
        ptr = llmemory.cast_int_to_adr(func).ptr
        if hasattr(ptr._obj, 'graph'):
            res = self.llinterp.eval_graph(ptr._obj.graph, args)
        else:
            res = ptr._obj._callable(*args)
        if RESULT is lltype.Void:
            return None
        return support.cast_result(RESULT, res)

    def _do_call(self, func, args_i, args_r, args_f, calldescr):
        TP = llmemory.cast_int_to_adr(func).ptr._obj._TYPE
        args = support.cast_call_args(TP.ARGS, args_i, args_r, args_f)
        return self.maybe_on_top_of_llinterp(func, args, TP.RESULT)

    bh_call_i = _do_call
    bh_call_r = _do_call
    bh_call_f = _do_call
    bh_call_v = _do_call

    def bh_getfield_gc(self, p, descr):
        p = support.cast_arg(lltype.Ptr(descr.S), p)
        return support.cast_result(descr.FIELD, getattr(p, descr.fieldname))

    bh_getfield_gc_i = bh_getfield_gc
    bh_getfield_gc_r = bh_getfield_gc
    bh_getfield_gc_f = bh_getfield_gc

    bh_getfield_raw = bh_getfield_gc
    bh_getfield_raw_i = bh_getfield_raw
    bh_getfield_raw_r = bh_getfield_raw
    bh_getfield_raw_f = bh_getfield_raw

    def bh_setfield_gc(self, p, newvalue, descr):
        p = support.cast_arg(lltype.Ptr(descr.S), p)
        setattr(p, descr.fieldname, support.cast_arg(descr.FIELD, newvalue))

    bh_setfield_gc_i = bh_setfield_gc
    bh_setfield_gc_r = bh_setfield_gc
    bh_setfield_gc_f = bh_setfield_gc

    bh_setfield_raw   = bh_setfield_gc
    bh_setfield_raw_i = bh_setfield_raw
    bh_setfield_raw_f = bh_setfield_raw

    def bh_arraylen_gc(self, a, descr):
        array = a._obj.container
        if descr.A is not descr.OUTERA:
            array = getattr(array, descr.OUTERA._arrayfld)
        return array.getlength()

    def bh_getarrayitem_gc(self, a, index, descr):
        a = support.cast_arg(lltype.Ptr(descr.A), a)
        array = a._obj
        assert index >= 0
        return support.cast_result(descr.A.OF, array.getitem(index))

    bh_getarrayitem_gc_pure_i = bh_getarrayitem_gc
    bh_getarrayitem_gc_pure_r = bh_getarrayitem_gc
    bh_getarrayitem_gc_pure_f = bh_getarrayitem_gc
    bh_getarrayitem_gc_i = bh_getarrayitem_gc
    bh_getarrayitem_gc_r = bh_getarrayitem_gc
    bh_getarrayitem_gc_f = bh_getarrayitem_gc

    bh_getarrayitem_raw = bh_getarrayitem_gc
    bh_getarrayitem_raw_i = bh_getarrayitem_raw
    bh_getarrayitem_raw_r = bh_getarrayitem_raw
    bh_getarrayitem_raw_f = bh_getarrayitem_raw

    def bh_setarrayitem_gc(self, a, index, item, descr):
        a = support.cast_arg(lltype.Ptr(descr.A), a)
        array = a._obj
        array.setitem(index, support.cast_arg(descr.A.OF, item))

    bh_setarrayitem_gc_i = bh_setarrayitem_gc
    bh_setarrayitem_gc_r = bh_setarrayitem_gc
    bh_setarrayitem_gc_f = bh_setarrayitem_gc

    bh_setarrayitem_raw   = bh_setarrayitem_gc
    bh_setarrayitem_raw_i = bh_setarrayitem_raw
    bh_setarrayitem_raw_r = bh_setarrayitem_raw
    bh_setarrayitem_raw_f = bh_setarrayitem_raw

    def bh_getinteriorfield_gc(self, a, index, descr):
        array = a._obj.container
        return support.cast_result(descr.FIELD,
                          getattr(array.getitem(index), descr.fieldname))

    bh_getinteriorfield_gc_i = bh_getinteriorfield_gc
    bh_getinteriorfield_gc_r = bh_getinteriorfield_gc
    bh_getinteriorfield_gc_f = bh_getinteriorfield_gc

    def bh_setinteriorfield_gc(self, a, index, item, descr):
        array = a._obj.container
        setattr(array.getitem(index), descr.fieldname,
                support.cast_arg(descr.FIELD, item))

    bh_setinteriorfield_gc_i = bh_setinteriorfield_gc
    bh_setinteriorfield_gc_r = bh_setinteriorfield_gc
    bh_setinteriorfield_gc_f = bh_setinteriorfield_gc

    def bh_raw_load_i(self, struct, offset, descr):
        ll_p = rffi.cast(rffi.CCHARP, struct)
        ll_p = rffi.cast(lltype.Ptr(descr.A), rffi.ptradd(ll_p, offset))
        value = ll_p[0]
        return support.cast_result(descr.A.OF, value)

    def bh_raw_load_f(self, struct, offset, descr):
        ll_p = rffi.cast(rffi.CCHARP, struct)
        ll_p = rffi.cast(rffi.CArrayPtr(longlong.FLOATSTORAGE),
                         rffi.ptradd(ll_p, offset))
        return ll_p[0]

    def bh_raw_load(self, struct, offset, descr):
        if descr.A.OF == lltype.Float:
            return self.bh_raw_load_f(struct, offset, descr)
        else:
            return self.bh_raw_load_i(struct, offset, descr)

    def _get_int_type_from_size(self, size):
        if   size == 1:
            return rffi.UCHAR
        elif size == 2:
            return rffi.USHORT
        elif size == 4:
            return rffi.UINT
        elif size == 8:
            return rffi.ULONGLONG
        elif size == -1:
            return rffi.SIGNEDCHAR
        elif size == -2:
            return rffi.SHORT
        elif size == -4:
            return rffi.INT
        elif size == -8:
            return rffi.LONGLONG
        else:
            raise NotImplementedError(size)

    def bh_gc_load_indexed_i(self, struct, index, scale, base_ofs, bytes):
        T = self._get_int_type_from_size(bytes)
        x = llop.gc_load_indexed(T, struct, index, scale, base_ofs)
        return lltype.cast_primitive(lltype.Signed, x)

    def bh_gc_load_indexed_f(self, struct, index, scale, base_ofs, bytes):
        if bytes != 8:
            raise Exception("gc_load_indexed_f is only for 'double'!")
        return llop.gc_load_indexed(longlong.FLOATSTORAGE,
                                    struct, index, scale, base_ofs)

    def bh_gc_store_indexed_i(self, struct, index, val, scale, base_ofs, bytes,
                              descr):
        T = self._get_int_type_from_size(bytes)
        val = lltype.cast_primitive(T, val)
        if descr.A.OF == lltype.SingleFloat:
            val = longlong.int2singlefloat(val)
        llop.gc_store_indexed(lltype.Void, struct, index, val, scale, base_ofs)

    def bh_gc_store_indexed_f(self, struct, index, val, scale, base_ofs, bytes,
                              descr):
        if bytes != 8:
            raise Exception("gc_store_indexed_f is only for 'double'!")
        val = longlong.getrealfloat(val)
        llop.gc_store_indexed(lltype.Void, struct, index, val, scale, base_ofs)

    def bh_gc_store_indexed(self, struct, index, val, scale, base_ofs, bytes,
                            descr):
        if descr.A.OF == lltype.Float:
            self.bh_gc_store_indexed_f(struct, index, val, scale, base_ofs,
                                       bytes, descr)
        else:
            self.bh_gc_store_indexed_i(struct, index, val, scale, base_ofs,
                                       bytes, descr)

    def bh_increment_debug_counter(self, addr):
        p = rffi.cast(rffi.CArrayPtr(lltype.Signed), addr)
        p[0] += 1

    def unpack_arraydescr_size(self, arraydescr):
        from rpython.jit.backend.llsupport.symbolic import get_array_token
        from rpython.jit.backend.llsupport.descr import get_type_flag, FLAG_SIGNED
        assert isinstance(arraydescr, ArrayDescr)
        basesize, itemsize, _ = get_array_token(arraydescr.A, False)
        flag = get_type_flag(arraydescr.A.OF)
        is_signed = (flag == FLAG_SIGNED)
        return basesize, itemsize, is_signed

    def bh_raw_store_i(self, struct, offset, newvalue, descr):
        ll_p = rffi.cast(rffi.CCHARP, struct)
        ll_p = rffi.cast(lltype.Ptr(descr.A), rffi.ptradd(ll_p, offset))
        if descr.A.OF == lltype.SingleFloat:
            newvalue = longlong.int2singlefloat(newvalue)
        ll_p[0] = rffi.cast(descr.A.OF, newvalue)

    def bh_raw_store_f(self, struct, offset, newvalue, descr):
        ll_p = rffi.cast(rffi.CCHARP, struct)
        ll_p = rffi.cast(rffi.CArrayPtr(longlong.FLOATSTORAGE),
                         rffi.ptradd(ll_p, offset))
        ll_p[0] = newvalue

    def bh_raw_store(self, struct, offset, newvalue, descr):
        if descr.A.OF == lltype.Float:
            self.bh_raw_store_f(struct, offset, newvalue, descr)
        else:
            self.bh_raw_store_i(struct, offset, newvalue, descr)

    def bh_newstr(self, length):
        return lltype.cast_opaque_ptr(llmemory.GCREF,
                                      lltype.malloc(rstr.STR, length,
                                                    zero=True))

    def bh_strlen(self, s):
        return s._obj.container.chars.getlength()

    def bh_strgetitem(self, s, item):
        assert item >= 0
        return ord(s._obj.container.chars.getitem(item))

    def bh_strsetitem(self, s, item, v):
        s._obj.container.chars.setitem(item, chr(v))

    def bh_copystrcontent(self, src, dst, srcstart, dststart, length):
        src = lltype.cast_opaque_ptr(lltype.Ptr(rstr.STR), src)
        dst = lltype.cast_opaque_ptr(lltype.Ptr(rstr.STR), dst)
        assert 0 <= srcstart <= srcstart + length <= len(src.chars)
        assert 0 <= dststart <= dststart + length <= len(dst.chars)
        rstr.copy_string_contents(src, dst, srcstart, dststart, length)

    def bh_strhash(self, s):
        lls = s._obj.container
        return compute_hash(hlstr(lls._as_ptr()))

    def bh_newunicode(self, length):
        return lltype.cast_opaque_ptr(llmemory.GCREF,
                                      lltype.malloc(rstr.UNICODE, length,
                                                    zero=True))

    def bh_unicodelen(self, string):
        return string._obj.container.chars.getlength()

    def bh_unicodegetitem(self, string, index):
        assert index >= 0
        return ord(string._obj.container.chars.getitem(index))

    def bh_unicodesetitem(self, string, index, newvalue):
        string._obj.container.chars.setitem(index, unichr(newvalue))

    def bh_copyunicodecontent(self, src, dst, srcstart, dststart, length):
        src = lltype.cast_opaque_ptr(lltype.Ptr(rstr.UNICODE), src)
        dst = lltype.cast_opaque_ptr(lltype.Ptr(rstr.UNICODE), dst)
        assert 0 <= srcstart <= srcstart + length <= len(src.chars)
        assert 0 <= dststart <= dststart + length <= len(dst.chars)
        rstr.copy_unicode_contents(src, dst, srcstart, dststart, length)

    def bh_unicodehash(self, s):
        lls = s._obj.container
        return compute_hash(hlunicode(lls._as_ptr()))

    def bh_new(self, sizedescr):
        return lltype.cast_opaque_ptr(llmemory.GCREF,
                                      lltype.malloc(sizedescr.S, zero=True))

    def bh_new_with_vtable(self, descr):
        result = lltype.malloc(descr.S, zero=True)
        result_as_objptr = lltype.cast_pointer(rclass.OBJECTPTR, result)
        result_as_objptr.typeptr = support.cast_from_int(rclass.CLASSTYPE,
                                                descr.get_vtable())
        return lltype.cast_opaque_ptr(llmemory.GCREF, result)

    def bh_new_array(self, length, arraydescr):
        array = lltype.malloc(arraydescr.A, length, zero=True)
        assert getkind(arraydescr.A.OF) != 'ref' # getkind crashes on structs
        return lltype.cast_opaque_ptr(llmemory.GCREF, array)

    def bh_new_array_clear(self, length, arraydescr):
        array = lltype.malloc(arraydescr.A, length, zero=True)
        return lltype.cast_opaque_ptr(llmemory.GCREF, array)

    def bh_classof(self, struct):
        struct = lltype.cast_opaque_ptr(rclass.OBJECTPTR, struct)
        return ptr2int(struct.typeptr)

    # vector operations
    vector_arith_code = """
    def bh_vec_{0}_{1}(self, vx, vy, count):
        assert len(vx) == len(vy) == count
        return [intmask(_vx {2} _vy) for _vx,_vy in zip(vx,vy)]
    """
    vector_float_arith_code = """
    def bh_vec_{0}_{1}(self, vx, vy, count):
        assert len(vx) == len(vy) == count
        return [_vx {2} _vy for _vx,_vy in zip(vx,vy)]
    """
    exec(py.code.Source(vector_arith_code.format('int','add','+')).compile())
    exec(py.code.Source(vector_arith_code.format('int','sub','-')).compile())
    exec(py.code.Source(vector_arith_code.format('int','mul','*')).compile())
    exec(py.code.Source(vector_arith_code.format('int','and','&')).compile())
    exec(py.code.Source(vector_arith_code.format('int','or','|')).compile())

    exec(py.code.Source(vector_float_arith_code.format('float','add','+')).compile())
    exec(py.code.Source(vector_float_arith_code.format('float','sub','-')).compile())
    exec(py.code.Source(vector_float_arith_code.format('float','mul','*')).compile())
    exec(py.code.Source(vector_float_arith_code.format('float','truediv','/')).compile())
    exec(py.code.Source(vector_float_arith_code.format('float','eq','==')).compile())

    def bh_vec_float_neg(self, vx, count):
        return [e * -1 for e in vx]

    def bh_vec_float_abs(self, vx, count):
        return [abs(e) for e in vx]

    def bh_vec_float_eq(self, vx, vy, count):
        assert len(vx) == len(vy) == count
        return [_vx == _vy for _vx,_vy in zip(vx,vy)]

    def bh_vec_float_ne(self, vx, vy, count):
        assert len(vx) == len(vy) == count
        return [_vx != _vy for _vx,_vy in zip(vx,vy)]

    bh_vec_int_eq = bh_vec_float_eq
    bh_vec_int_ne = bh_vec_float_ne

    def bh_vec_int_is_true(self, vx, count):
        return map(lambda x: bool(x), vx)

    def bh_vec_int_is_false(self, vx, count):
        return map(lambda x: not bool(x), vx)

    def bh_vec_int_xor(self, vx, vy, count):
        return [int(x) ^ int(y) for x,y in zip(vx,vy)]

    def bh_vec_float_xor(self, vx, vy, count):
        return [0.0 for x,y in zip(vx,vy)] # just used for clearing the vector register

    def bh_vec_cast_float_to_singlefloat(self, vx, count):
        from rpython.rlib.rarithmetic import r_singlefloat
        return [longlong.singlefloat2int(r_singlefloat(longlong.getrealfloat(v)))
                for v in vx]

    def bh_vec_cast_singlefloat_to_float(self, vx, count):
        return [longlong.getfloatstorage(float(longlong.int2singlefloat(v)))
                for v in vx]

    def bh_vec_cast_float_to_int(self, vx, count):
        return [int(x) for x in vx]

    def bh_vec_cast_int_to_float(self, vx, count):
        return [float(x) for x in vx]

    def bh_vec_f(self, count):
        return [0.0] * count

    def bh_vec_i(self, count):
        return [0] * count

    def _bh_vec_pack(self, tv, sv, index, count, newcount):
        while len(tv) < newcount: tv.append(None)
        if not isinstance(sv, list):
            tv[index] = sv
            return tv
        for i in range(count):
            tv[index+i] = sv[i]
        return tv

    bh_vec_pack_f = _bh_vec_pack
    bh_vec_pack_i = _bh_vec_pack

    def _bh_vec_unpack(self, vx, index, count, newcount):
        return vx[index:index+count]

    bh_vec_unpack_f = _bh_vec_unpack
    bh_vec_unpack_i = _bh_vec_unpack

    def _bh_vec_expand(self, x, count):
        return [x] * count

    bh_vec_expand_f = _bh_vec_expand
    bh_vec_expand_i = _bh_vec_expand

    def bh_vec_int_signext(self, vx, ext, count):
        return [int_signext(_vx, ext) for _vx in vx]

    def build_load(func):
        def load(self, struct, offset, scale, disp, descr, _count):
            values = []
            count = self.vector_ext.vec_size() // descr.get_item_size_in_bytes()
            assert _count == count
            assert count > 0
            adr = support.addr_add_bytes(struct, (offset * scale + disp))
            a = support.cast_arg(lltype.Ptr(descr.A), adr)
            array = a._obj
            for i in range(count):
                val = support.cast_result(descr.A.OF, array.getitem(i))
                values.append(val)
            return values
        return load

    bh_vec_load_i = build_load(bh_getarrayitem_raw)
    bh_vec_load_f = build_load(bh_getarrayitem_raw)
    del build_load

    def bh_vec_store(self, struct, offset, newvalues, scale, disp, descr, count):
        adr = support.addr_add_bytes(struct, offset * scale + disp)
        a = support.cast_arg(lltype.Ptr(descr.A), adr)
        array = a._obj
        for i,n in enumerate(newvalues):
            array.setitem(i, support.cast_arg(descr.A.OF, n))

    def store_fail_descr(self, deadframe, descr):
        pass # I *think*

    def protect_speculative_field(self, p, fielddescr):
        if not p:
            raise SpeculativeError
        p = p._obj.container._as_ptr()
        try:
            lltype.cast_pointer(lltype.Ptr(fielddescr.S), p)
        except lltype.InvalidCast:
            raise SpeculativeError

    def protect_speculative_array(self, p, arraydescr):
        if not p:
            raise SpeculativeError
        p = p._obj.container
        if lltype.typeOf(p) != arraydescr.A:
            raise SpeculativeError

    def protect_speculative_string(self, p):
        if not p:
            raise SpeculativeError
        p = p._obj.container
        if lltype.typeOf(p) != rstr.STR:
            raise SpeculativeError

    def protect_speculative_unicode(self, p):
        if not p:
            raise SpeculativeError
        p = p._obj.container
        if lltype.typeOf(p) != rstr.UNICODE:
            raise SpeculativeError


class LLDeadFrame(object):
    _TYPE = llmemory.GCREF

    def __init__(self, latest_descr, values,
                 last_exception=None, saved_data=None,
                 extra_value=None):
        self._latest_descr = latest_descr
        self._values = values
        self._last_exception = last_exception
        self._saved_data = saved_data
        self._extra_value = extra_value


class LLFrame(object):
    _TYPE = llmemory.GCREF

    forced_deadframe = None
    overflow_flag = False
    last_exception = None
    force_guard_op = None

    def __init__(self, cpu, argboxes, args):
        self.env = {}
        self.cpu = cpu
        assert len(argboxes) == len(args)
        for box, arg in zip(argboxes, args):
            self.setenv(box, arg)

    def __eq__(self, other):
        # this is here to avoid crashes in 'token == TOKEN_TRACING_RESCALL'
        from rpython.jit.metainterp.virtualizable import TOKEN_NONE
        from rpython.jit.metainterp.virtualizable import TOKEN_TRACING_RESCALL
        if isinstance(other, LLFrame):
            return self is other
        if other == TOKEN_NONE or other == TOKEN_TRACING_RESCALL:
            return False
        assert 0

    def __ne__(self, other):
        return not (self == other)

    def _identityhash(self):
        return hash(self)

    def setenv(self, box, arg):
        if box.is_vector() and box.count > 1:
            if box.datatype == INT:
                for i,a in enumerate(arg):
                    if isinstance(a, bool):
                        arg[i] = int(a)
                assert all([lltype.typeOf(a) == lltype.Signed for a in arg])
            elif box.datatype == FLOAT:
                assert all([lltype.typeOf(a) == longlong.FLOATSTORAGE or \
                            lltype.typeOf(a) == lltype.Signed for a in arg])
            else:
                raise AssertionError(box)
        elif box.type == INT:
            # typecheck the result
            if isinstance(arg, bool):
                arg = int(arg)
            assert lltype.typeOf(arg) == lltype.Signed
        elif box.type == REF:
            assert lltype.typeOf(arg) == llmemory.GCREF
        elif box.type == FLOAT:
            assert lltype.typeOf(arg) == longlong.FLOATSTORAGE
        else:
            raise AssertionError(box)
        #
        self.env[box] = arg

    def lookup(self, arg):
        if isinstance(arg, Const):
            return arg.value
        return self.env[arg]

    def execute(self, lltrace):
        self.lltrace = lltrace
        del lltrace
        i = 0
        while True:
            assert not self.lltrace.has_been_freed
            op = self.lltrace.operations[i]
            args = [self.lookup(arg) for arg in op.getarglist()]
            self.current_op = op # for label
            self.current_index = i
            execute = getattr(self, 'execute_' + op.getopname())
            try:
                resval = execute(_getdescr(op), *args)
            except Jump as j:
                self.lltrace, i = j.jump_target
                if i >= 0:
                    label_op = self.lltrace.operations[i]
                    i += 1
                    targetargs = label_op.getarglist()
                else:
                    targetargs = self.lltrace.inputargs
                    i = 0
                self.do_renaming(targetargs, j.args)
                continue
            if op.type != 'v':
                self.setenv(op, resval)
            else:
                assert resval is None
            i += 1

    def do_renaming(self, newargs, newvalues):
        assert len(newargs) == len(newvalues)
        self.env = {}
        self.framecontent = {}
        for new, newvalue in zip(newargs, newvalues):
            self.setenv(new, newvalue)

    # -----------------------------------------------------

    def _accumulate(self, descr, failargs, values):
        info = descr.rd_vector_info
        while info:
            i = info.getpos_in_failargs()
            value = values[i]
            assert isinstance(value, list)
            if info.accum_operation == '+':
                value = sum(value)
            elif info.accum_operation == '*':
                def prod(acc, x): return acc * x
                value = reduce(prod, value, 1.0)
            else:
                raise NotImplementedError("accum operator in fail guard")
            values[i] = value
            info = info.next()

    def fail_guard(self, descr, saved_data=None, extra_value=None,
                   propagate_exception=False):
        if not propagate_exception:
            assert self.last_exception is None
        values = []
        for box in self.current_op.getfailargs():
            if box is not None:
                value = self.env[box]
            else:
                value = None
            values.append(value)
        self._accumulate(descr, self.current_op.getfailargs(), values)
        if hasattr(descr, '_llgraph_bridge'):
            if propagate_exception:
                assert (descr._llgraph_bridge.operations[0].opnum in
                        (rop.SAVE_EXC_CLASS, rop.GUARD_EXCEPTION,
                         rop.GUARD_NO_EXCEPTION))
            target = (descr._llgraph_bridge, -1)
            values = [value for value in values if value is not None]
            raise Jump(target, values)
        else:
            raise ExecutionFinished(LLDeadFrame(descr, values,
                                                self.last_exception,
                                                saved_data, extra_value))

    def execute_force_spill(self, _, arg):
        pass

    def execute_finish(self, descr, *args):
        raise ExecutionFinished(LLDeadFrame(descr, args))

    def execute_label(self, descr, *args):
        argboxes = self.current_op.getarglist()
        self.do_renaming(argboxes, args)

    def _test_true(self, arg):
        assert arg in (0, 1)
        return arg

    def _test_false(self, arg):
        assert arg in (0, 1)
        return arg

    def execute_vec_guard_true(self, descr, arg):
        assert isinstance(arg, list)
        if not all(arg):
            self.fail_guard(descr)

    def execute_vec_guard_false(self, descr, arg):
        assert isinstance(arg, list)
        if any(arg):
            self.fail_guard(descr)

    def execute_guard_true(self, descr, arg):
        if not self._test_true(arg):
            self.fail_guard(descr)

    def execute_guard_false(self, descr, arg):
        if self._test_false(arg):
            self.fail_guard(descr)

    def execute_guard_value(self, descr, arg1, arg2):
        if arg1 != arg2:
            self.fail_guard(descr, extra_value=arg1)

    def execute_guard_nonnull(self, descr, arg):
        if not arg:
            self.fail_guard(descr)

    def execute_guard_isnull(self, descr, arg):
        if arg:
            self.fail_guard(descr)

    def execute_guard_class(self, descr, arg, klass):
        value = lltype.cast_opaque_ptr(rclass.OBJECTPTR, arg)
        expected_class = llmemory.cast_adr_to_ptr(
            llmemory.cast_int_to_adr(klass),
            rclass.CLASSTYPE)
        if value.typeptr != expected_class:
            self.fail_guard(descr)

    def execute_guard_nonnull_class(self, descr, arg, klass):
        self.execute_guard_nonnull(descr, arg)
        self.execute_guard_class(descr, arg, klass)

    def execute_guard_gc_type(self, descr, arg, typeid):
        assert isinstance(typeid, TypeIDSymbolic)
        TYPE = arg._obj.container._TYPE
        if TYPE != typeid.STRUCT_OR_ARRAY:
            self.fail_guard(descr)

    def execute_guard_is_object(self, descr, arg):
        TYPE = arg._obj.container._TYPE
        while TYPE is not rclass.OBJECT:
            if not isinstance(TYPE, lltype.GcStruct):   # or TYPE is None
                self.fail_guard(descr)
                return
            _, TYPE = TYPE._first_struct()

    def execute_guard_subclass(self, descr, arg, klass):
        value = lltype.cast_opaque_ptr(rclass.OBJECTPTR, arg)
        expected_class = llmemory.cast_adr_to_ptr(
            llmemory.cast_int_to_adr(klass),
            rclass.CLASSTYPE)
        if (expected_class.subclassrange_min
                <= value.typeptr.subclassrange_min
                <= expected_class.subclassrange_max):
            pass
        else:
            self.fail_guard(descr)

    def execute_guard_no_exception(self, descr):
        if self.last_exception is not None:
            self.fail_guard(descr, propagate_exception=True)

    def execute_guard_exception(self, descr, excklass):
        lle = self.last_exception
        if lle is None:
            gotklass = lltype.nullptr(rclass.CLASSTYPE.TO)
        else:
            gotklass = lle.args[0]
        excklass = llmemory.cast_adr_to_ptr(
            llmemory.cast_int_to_adr(excklass),
            rclass.CLASSTYPE)
        if gotklass != excklass:
            self.fail_guard(descr, propagate_exception=True)
        #
        res = lle.args[1]
        self.last_exception = None
        return support.cast_to_ptr(res)

    def execute_guard_not_forced(self, descr):
        if self.forced_deadframe is not None:
            saved_data = self.forced_deadframe._saved_data
            self.fail_guard(descr, saved_data, propagate_exception=True)
        self.force_guard_op = self.current_op
    execute_guard_not_forced_2 = execute_guard_not_forced

    def execute_guard_not_invalidated(self, descr):
        if self.lltrace.invalid:
            self.fail_guard(descr)

    def execute_guard_always_fails(self, descr):
        self.fail_guard(descr)

    def execute_int_add_ovf(self, _, x, y):
        try:
            z = ovfcheck(x + y)
        except OverflowError:
            ovf = True
            z = 0
        else:
            ovf = False
        self.overflow_flag = ovf
        return z

    def execute_int_sub_ovf(self, _, x, y):
        try:
            z = ovfcheck(x - y)
        except OverflowError:
            ovf = True
            z = 0
        else:
            ovf = False
        self.overflow_flag = ovf
        return z

    def execute_int_mul_ovf(self, _, x, y):
        try:
            z = ovfcheck(x * y)
        except OverflowError:
            ovf = True
            z = 0
        else:
            ovf = False
        self.overflow_flag = ovf
        return z

    def execute_guard_no_overflow(self, descr):
        if self.overflow_flag:
            self.fail_guard(descr)

    def execute_guard_overflow(self, descr):
        if not self.overflow_flag:
            self.fail_guard(descr)

    def execute_jump(self, descr, *args):
        raise Jump(descr._llgraph_target, args)

    def _do_math_sqrt(self, value):
        import math
        y = support.cast_from_floatstorage(lltype.Float, value)
        x = math.sqrt(y)
        return support.cast_to_floatstorage(x)

    def execute_cond_call(self, calldescr, cond, func, *args):
        if not cond:
            return
        # cond_call can't have a return value
        self.execute_call_n(calldescr, func, *args)

    def execute_cond_call_value_i(self, calldescr, value, func, *args):
        if not value:
            value = self.execute_call_i(calldescr, func, *args)
        return value

    def execute_cond_call_value_r(self, calldescr, value, func, *args):
        if not value:
            value = self.execute_call_r(calldescr, func, *args)
        return value

    def _execute_call(self, calldescr, func, *args):
        effectinfo = calldescr.get_extra_info()
        if effectinfo is not None and hasattr(effectinfo, 'oopspecindex'):
            oopspecindex = effectinfo.oopspecindex
            if oopspecindex == EffectInfo.OS_MATH_SQRT:
                return self._do_math_sqrt(args[0])
        TP = llmemory.cast_int_to_adr(func).ptr._obj._TYPE
        call_args = support.cast_call_args_in_order(TP.ARGS, args)
        try:
            res = self.cpu.maybe_on_top_of_llinterp(func, call_args, TP.RESULT)
            self.last_exception = None
        except LLException as lle:
            self.last_exception = lle
            res = _example_res[getkind(TP.RESULT)[0]]
        return res

    execute_call_i = _execute_call
    execute_call_r = _execute_call
    execute_call_f = _execute_call
    execute_call_n = _execute_call

    def _execute_call_may_force(self, calldescr, func, *args):
        guard_op = self.lltrace.operations[self.current_index + 1]
        assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
        self.force_guard_op = guard_op
        res = self._execute_call(calldescr, func, *args)
        del self.force_guard_op
        return res

    execute_call_may_force_n = _execute_call_may_force
    execute_call_may_force_r = _execute_call_may_force
    execute_call_may_force_f = _execute_call_may_force
    execute_call_may_force_i = _execute_call_may_force

    def _execute_call_release_gil(self, descr, saveerr, func, *args):
        if hasattr(descr, '_original_func_'):
            func = descr._original_func_     # see pyjitpl.py
            # we want to call the function that does the aroundstate
            # manipulation here (as a hack, instead of really doing
            # the aroundstate manipulation ourselves)
            return self._execute_call_may_force(descr, func, *args)
        guard_op = self.lltrace.operations[self.current_index + 1]
        assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
        self.force_guard_op = guard_op
        call_args = support.cast_call_args_in_order(descr.ARGS, args)
        #
        func_adr = llmemory.cast_int_to_adr(func)
        if hasattr(func_adr.ptr._obj, '_callable'):
            # this is needed e.g. by test_fficall.test_guard_not_forced_fails,
            # because to actually force the virtualref we need to llinterp the
            # graph, not to directly execute the python function
            result = self.cpu.maybe_on_top_of_llinterp(func, call_args, descr.RESULT)
        else:
            FUNC = lltype.FuncType(descr.ARGS, descr.RESULT, descr.ABI)
            func_to_call = rffi.cast(lltype.Ptr(FUNC), func)
            result = func_to_call(*call_args)
        del self.force_guard_op
        return support.cast_result(descr.RESULT, result)

    execute_call_release_gil_n = _execute_call_release_gil
    execute_call_release_gil_i = _execute_call_release_gil
    execute_call_release_gil_f = _execute_call_release_gil

    def _new_execute_call_assembler(def_val):
        def _execute_call_assembler(self, descr, *args):
            # XXX simplify the following a bit
            #
            # pframe = CALL_ASSEMBLER(args..., descr=looptoken)
            # ==>
            #     pframe = CALL looptoken.loopaddr(*args)
            #     JUMP_IF_FAST_PATH @fastpath
            #     res = CALL assembler_call_helper(pframe)
            #     jmp @done
            #   @fastpath:
            #     res = GETFIELD(pframe, 'result')
            #   @done:
            #
            call_op = self.lltrace.operations[self.current_index]
            guard_op = self.lltrace.operations[self.current_index + 1]
            assert guard_op.getopnum() == rop.GUARD_NOT_FORCED
            self.force_guard_op = guard_op
            pframe = self.cpu._execute_token(descr, *args)
            del self.force_guard_op
            #
            jd = descr.outermost_jitdriver_sd
            assert jd is not None, ("call_assembler(): the loop_token needs "
                                    "to have 'outermost_jitdriver_sd'")
            if jd.index_of_virtualizable != -1:
                vable = args[jd.index_of_virtualizable]
            else:
                vable = lltype.nullptr(llmemory.GCREF.TO)
            #
            # Emulate the fast path
            #
            faildescr = self.cpu.get_latest_descr(pframe)
            if faildescr == self.cpu.done_with_this_frame_descr_int:
                return self.cpu.get_int_value(pframe, 0)
            elif faildescr == self.cpu.done_with_this_frame_descr_ref:
                return self.cpu.get_ref_value(pframe, 0)
            elif faildescr == self.cpu.done_with_this_frame_descr_float:
                return self.cpu.get_float_value(pframe, 0)
            elif faildescr == self.cpu.done_with_this_frame_descr_void:
                return None

            assembler_helper_ptr = jd.assembler_helper_adr.ptr  # fish
            try:
                result = assembler_helper_ptr(pframe, vable)
            except LLException as lle:
                assert self.last_exception is None, "exception left behind"
                self.last_exception = lle
                # fish op
                result = def_val
            if isinstance(result, float):
                result = support.cast_to_floatstorage(result)
            return result
        return _execute_call_assembler

    execute_call_assembler_i = _new_execute_call_assembler(0)
    execute_call_assembler_r = _new_execute_call_assembler(lltype.nullptr(llmemory.GCREF.TO))
    execute_call_assembler_f = _new_execute_call_assembler(0.0)
    execute_call_assembler_n = _new_execute_call_assembler(None)

    def execute_same_as_i(self, _, x):
        return x
    execute_same_as_f = execute_same_as_i
    execute_same_as_r = execute_same_as_i

    def execute_debug_merge_point(self, descr, *args):
        from rpython.jit.metainterp.warmspot import get_stats
        try:
            stats = get_stats()
        except AttributeError:
            pass
        else:
            stats.add_merge_point_location(args[1:])

    def execute_enter_portal_frame(self, descr, *args):
        pass

    def execute_leave_portal_frame(self, descr, *args):
        pass

    def execute_new_with_vtable(self, descr):
        return self.cpu.bh_new_with_vtable(descr)

    def execute_force_token(self, _):
        return self

    def execute_cond_call_gc_wb(self, descr, a):
        py.test.skip("cond_call_gc_wb not supported")

    def execute_cond_call_gc_wb_array(self, descr, a, b):
        py.test.skip("cond_call_gc_wb_array not supported")

    def execute_keepalive(self, descr, x):
        pass

    def execute_save_exc_class(self, descr):
        lle = self.last_exception
        if lle is None:
            return 0
        else:
            return support.cast_to_int(lle.args[0])

    def execute_save_exception(self, descr):
        lle = self.last_exception
        if lle is None:
            res = lltype.nullptr(llmemory.GCREF.TO)
        else:
            res = lltype.cast_opaque_ptr(llmemory.GCREF, lle.args[1])
        self.last_exception = None
        return res

    def execute_restore_exception(self, descr, kls, e):
        if e:
            value = lltype.cast_opaque_ptr(rclass.OBJECTPTR, e)
            assert ptr2int(value.typeptr) == kls
            lle = LLException(value.typeptr, e)
        else:
            assert kls == 0
            lle = None
        self.last_exception = lle

    def execute_check_memory_error(self, descr, value):
        if not value:
            from rpython.jit.backend.llsupport import llmodel
            raise llmodel.MissingLatestDescrError


def _getdescr(op):
    d = op.getdescr()
    if d is not None and isinstance(d, WeakrefDescr):
        d = d.realdescrref()
        assert d is not None, "the descr disappeared: %r" % (op,)
    return d

def _setup():
    def _make_impl_from_blackhole_interp(opname):
        from rpython.jit.metainterp.blackhole import BlackholeInterpreter
        name = 'bhimpl_' + opname.lower()
        try:
            func = BlackholeInterpreter.__dict__[name]
        except KeyError:
            return
        for argtype in func.argtypes:
            if argtype not in ('i', 'r', 'f'):
                return
        #
        def _op_default_implementation(self, descr, *args):
            # for all operations implemented in the blackhole interpreter
            return func(*args)
        #
        _op_default_implementation.__name__ = 'execute_' + opname
        return _op_default_implementation

    def _new_execute(opname):
        def execute(self, descr, *args):
            if descr is not None:
                new_args = args + (descr,)
            else:
                new_args = args
            if opname.startswith('vec_'):
                # pre vector op
                count = self.current_op.count
                assert count >= 0
                new_args = new_args + (count,)
            result = getattr(self.cpu, 'bh_' + opname)(*new_args)
            if isinstance(result, list):
                # post vector op
                count = self.current_op.count
                if len(result) > count:
                    assert count > 0
                    result = result[:count]
                if count == 1:
                    result = result[0]
            return result
        execute.__name__ = 'execute_' + opname
        return execute

    for k, v in rop.__dict__.iteritems():
        if not k.startswith("_"):
            fname = 'execute_' + k.lower()
            if not hasattr(LLFrame, fname):
                func = _make_impl_from_blackhole_interp(k)
                if func is None:
                    func = _new_execute(k.lower())
                setattr(LLFrame, fname, func)

_setup()



class AppTestCProfile(object):
    spaceconfig = {
        "usemodules": ['_lsprof', 'time'],
    }

    def setup_class(cls):
        cls.w_file = cls.space.wrap(__file__)

    def test_repr(self):
        import _lsprof
        assert repr(_lsprof.Profiler) == "<type '_lsprof.Profiler'>"

    def test_builtins(self):
        import _lsprof
        prof = _lsprof.Profiler()
        lst = []
        prof.enable()
        lst.append(len(lst))
        prof.disable()
        stats = prof.getstats()
        expected = (
            "<len>",
            "<method 'append' of 'list' objects>",
            "<method 'disable' of '_lsprof.Profiler' objects>",
        )
        for entry in stats:
            assert entry.code in expected

    def test_builtins_callers(self):
        import _lsprof
        prof = _lsprof.Profiler(subcalls=True)
        lst = []
        def f1():
            lst.append(len(lst))
        prof.enable(subcalls=True)
        f1()
        prof.disable()
        stats = prof.getstats()
        expected = (
            "<len>",
            "<method 'append' of 'list' objects>",
        )
        by_id = set()
        for entry in stats:
            if entry.code == f1.__code__:
                assert len(entry.calls) == 2
                for subentry in entry.calls:
                    assert subentry.code in expected
                    by_id.add(id(subentry.code))
            elif entry.code in expected:
                by_id.add(id(entry.code))
        #  :-(  cProfile.py relies on the id() of the strings...
        assert len(by_id) == len(expected)

    def test_direct(self):
        import _lsprof
        def getticks():
            return len(ticks)
        prof = _lsprof.Profiler(getticks, 0.25, True, False)
        ticks = []
        def bar(m):
            ticks.append(1)
            if m == 1:
                foo(42)
            ticks.append(1)
        def spam(m):
            bar(m)
        def foo(n):
            bar(n)
            ticks.append(1)
            bar(n+1)
            ticks.append(1)
            spam(n+2)
        prof.enable()
        foo(0)
        prof.disable()
        assert len(ticks) == 16
        stats = prof.getstats()
        entries = {}
        for entry in stats:
            assert hasattr(entry.code, 'co_name')
            entries[entry.code.co_name] = entry
        efoo = entries['foo']
        assert efoo.callcount == 2
        assert efoo.reccallcount == 1
        assert efoo.inlinetime == 1.0
        assert efoo.totaltime == 4.0
        assert len(efoo.calls) == 2
        ebar = entries['bar']
        assert ebar.callcount == 6
        assert ebar.reccallcount == 3
        assert ebar.inlinetime == 3.0
        assert ebar.totaltime == 3.5
        assert len(ebar.calls) == 1
        espam = entries['spam']
        assert espam.callcount == 2
        assert espam.reccallcount == 0
        assert espam.inlinetime == 0.0
        assert espam.totaltime == 1.0
        assert len(espam.calls) == 1

        foo2spam, foo2bar = efoo.calls
        if foo2bar.code.co_name == 'spam':
            foo2bar, foo2spam = foo2spam, foo2bar
        assert foo2bar.code.co_name == 'bar'
        assert foo2bar.callcount == 4
        assert foo2bar.reccallcount == 2
        assert foo2bar.inlinetime == 2.0
        assert foo2bar.totaltime == 3.0
        assert foo2spam.code.co_name == 'spam'
        assert foo2spam.callcount == 2
        assert foo2spam.reccallcount == 0
        assert foo2spam.inlinetime == 0.0
        assert foo2spam.totaltime == 1.0

        bar2foo, = ebar.calls
        assert bar2foo.code.co_name == 'foo'
        assert bar2foo.callcount == 1
        assert bar2foo.reccallcount == 0
        assert bar2foo.inlinetime == 0.5
        assert bar2foo.totaltime == 2.0

        spam2bar, = espam.calls
        assert spam2bar.code.co_name == 'bar'
        assert spam2bar.callcount == 2
        assert spam2bar.reccallcount == 0
        assert spam2bar.inlinetime == 1.0
        assert spam2bar.totaltime == 1.0

    def test_scale_of_result(self):
        import _lsprof, time
        prof = _lsprof.Profiler()
        def foo(n):
            t = time.time()
            while abs(t - time.time()) < 1.0:
                pass      # busy-wait for 1 second
        def bar(n):
            foo(n)
        prof.enable()
        bar(0)
        prof.disable()
        stats = prof.getstats()
        entries = {}
        for entry in stats:
            entries[entry.code] = entry
        efoo = entries[foo.__code__]
        ebar = entries[bar.__code__]
        assert 0.9 < efoo.totaltime < 2.9
        # --- cannot test .inlinetime, because it does not include
        # --- the time spent doing the call to time.time()
        #assert 0.9 < efoo.inlinetime < 2.9
        for subentry in ebar.calls:
            assert 0.9 < subentry.totaltime < 2.9
            #assert 0.9 < subentry.inlinetime < 2.9

    def test_builtin_exception(self):
        import math
        import _lsprof

        prof = _lsprof.Profiler()
        prof.enable()
        try:
            math.sqrt("a")
        except TypeError:
            pass
        prof.disable()
        stats = prof.getstats()
        assert len(stats) == 2



from rpython.rtyper.lltypesystem import lltype, llmemory, rffi
from rpython.rtyper.annlowlevel import llhelper
from rpython.rlib.objectmodel import specialize
from rpython.rlib.debug import ll_assert
from rpython.rlib.objectmodel import enforceargs
from rpython.rlib import rgc

SIZEOFSIGNED = rffi.sizeof(lltype.Signed)
IS_32BIT = (SIZEOFSIGNED == 4)

# this is an info that only depends on the assembler executed, copied from
# compiled loop token (in fact we could use this as a compiled loop token
# XXX do this

GCMAP = lltype.Array(lltype.Unsigned)
NULLGCMAP = lltype.nullptr(GCMAP)

@enforceargs(None, int, int)
def jitframeinfo_update_depth(jfi, base_ofs, new_depth):
    if new_depth > jfi.jfi_frame_depth:
        jfi.jfi_frame_depth = new_depth
        jfi.jfi_frame_size = base_ofs + new_depth * SIZEOFSIGNED

def jitframeinfo_clear(jfi):
    jfi.jfi_frame_size = 0
    jfi.jfi_frame_depth = 0

JITFRAMEINFO_SIZE = 2 * SIZEOFSIGNED # make sure this stays correct

JITFRAMEINFO = lltype.Struct(
    'JITFRAMEINFO',
    # the depth of the frame
    ('jfi_frame_depth', lltype.Signed),
    # the total size of the frame, in bytes
    ('jfi_frame_size', lltype.Signed),
    adtmeths = {
        'update_frame_depth': jitframeinfo_update_depth,
        'clear': jitframeinfo_clear,
    },
)

NULLFRAMEINFO = lltype.nullptr(JITFRAMEINFO)
JITFRAMEINFOPTR = lltype.Ptr(JITFRAMEINFO)

# the JITFRAME that's stored on the heap. See backend/<backend>/arch.py for
# detailed explanation how it is on your architecture

def jitframe_allocate(frame_info):
    rgc.register_custom_trace_hook(JITFRAME, lambda_jitframe_trace)
    frame = lltype.malloc(JITFRAME, frame_info.jfi_frame_depth)
    frame.jf_frame_info = frame_info
    return frame

def jitframe_resolve(frame):
    while frame.jf_forward:
        frame = frame.jf_forward
    return frame

JITFRAME = lltype.GcForwardReference()

JITFRAME.become(lltype.GcStruct(
    'JITFRAME',
    ('jf_frame_info', lltype.Ptr(JITFRAMEINFO)),
    # Once the execute_token() returns, the field 'jf_descr' stores the
    # descr of the last executed operation (either a GUARD, or FINISH).
    # This field is also set immediately before doing CALL_MAY_FORCE
    # or CALL_ASSEMBLER.
    ('jf_descr', llmemory.GCREF),
    # guard_not_forced descr
    ('jf_force_descr', llmemory.GCREF),
    # a map of GC pointers
    ('jf_gcmap', lltype.Ptr(GCMAP)),
    # For the front-end: a GCREF for the savedata
    ('jf_savedata', llmemory.GCREF),
    # For GUARD_(NO)_EXCEPTION and GUARD_NOT_FORCED: the exception we
    # got.  (Note that in case of a regular FINISH generated from
    # RPython code that finishes the function with an exception, the
    # exception is not stored there, but is simply kept as a variable there)
    ('jf_guard_exc', llmemory.GCREF),
    # in case the frame got reallocated, we have to forward it somewhere
    ('jf_forward', lltype.Ptr(JITFRAME)),
    # the actual frame
    ('jf_frame', lltype.Array(lltype.Signed)),
    # note that we keep length field, because it's crucial to have the data
    # about GCrefs here and not in frame info which might change
    adtmeths = {
        'allocate': jitframe_allocate,
        'resolve': jitframe_resolve,
    },
    rtti = True,
))

@specialize.memo()
def getofs(name):
    return llmemory.offsetof(JITFRAME, name)

GCMAPLENGTHOFS = llmemory.arraylengthoffset(GCMAP)
GCMAPBASEOFS = llmemory.itemoffsetof(GCMAP, 0)
BASEITEMOFS = llmemory.itemoffsetof(JITFRAME.jf_frame, 0)
LENGTHOFS = llmemory.arraylengthoffset(JITFRAME.jf_frame)
SIGN_SIZE = llmemory.sizeof(lltype.Signed)
UNSIGN_SIZE = llmemory.sizeof(lltype.Unsigned)

def jitframe_trace(gc, obj_addr, callback, arg1, arg2):
    gc._trace_callback(callback, arg1, arg2, obj_addr + getofs('jf_descr'))
    gc._trace_callback(callback, arg1, arg2,obj_addr + getofs('jf_force_descr'))
    gc._trace_callback(callback, arg1, arg2, obj_addr + getofs('jf_savedata'))
    gc._trace_callback(callback, arg1, arg2, obj_addr + getofs('jf_guard_exc'))
    gc._trace_callback(callback, arg1, arg2, obj_addr + getofs('jf_forward'))

    if IS_32BIT:
        MAX = 32
    else:
        MAX = 64
    gcmap = (obj_addr + getofs('jf_gcmap')).address[0]
    if not gcmap:
        return      # done
    gcmap_lgt = (gcmap + GCMAPLENGTHOFS).signed[0]
    no = 0
    while no < gcmap_lgt:
        cur = (gcmap + GCMAPBASEOFS + UNSIGN_SIZE * no).unsigned[0]
        bitindex = 0
        while bitindex < MAX:
            if cur & (1 << bitindex):
                # the 'bitindex' is set in 'cur'
                index = no * SIZEOFSIGNED * 8 + bitindex
                # sanity check
                frame_lgt = (obj_addr + getofs('jf_frame') + LENGTHOFS) \
                    .signed[0]
                ll_assert(index < frame_lgt, "bogus frame field get")
                gc._trace_callback(callback, arg1, arg2,
                                   obj_addr + getofs('jf_frame') +
                                   BASEITEMOFS + SIGN_SIZE * index)
            bitindex += 1
        no += 1
lambda_jitframe_trace = lambda: jitframe_trace

JITFRAMEPTR = lltype.Ptr(JITFRAME)



import os
import sys

import py

from rpython.config.config import (OptionDescription, BoolOption, IntOption,
  ChoiceOption, StrOption, to_optparse)
from rpython.config.translationoption import IS_64_BITS


modulepath = py.path.local(__file__).dirpath().dirpath().join("module")
all_modules = [p.basename for p in modulepath.listdir()
               if p.check(dir=True, dotfile=False)
               and p.join('__init__.py').check()
               and not p.basename.startswith('test')]

essential_modules = set([
    "exceptions", "_file", "sys", "__builtin__", "posix", "_warnings",
    "itertools"
])

default_modules = essential_modules.copy()
default_modules.update([
    "_codecs", "gc", "_weakref", "marshal", "errno", "imp", "math", "cmath",
    "_sre", "_pickle_support", "operator", "parser", "symbol", "token", "_ast",
    "_io", "_random", "__pypy__", "_testing", "time"
])


# --allworkingmodules
working_modules = default_modules.copy()
working_modules.update([
    "_socket", "unicodedata", "mmap", "fcntl", "_locale", "pwd",
    "select", "zipimport", "_lsprof", "signal", "_rawffi", "termios",
    "zlib", "bz2", "struct", "_md5", "_sha", "_minimal_curses",
    "cStringIO", "thread", "itertools", "pyexpat", "cpyext", "array",
    "binascii", "_multiprocessing", '_warnings', "_collections",
    "_multibytecodec", "micronumpy", "_continuation", "_cffi_backend",
    "_csv", "_cppyy", "_pypyjson", "_jitlog",
    # "_hashlib", "crypt"
])

import rpython.rlib.rvmprof.cintf
if rpython.rlib.rvmprof.cintf.IS_SUPPORTED:
    working_modules.add('_vmprof')
    working_modules.add('faulthandler')

translation_modules = default_modules.copy()
translation_modules.update([
    "fcntl", "time", "select", "signal", "_rawffi", "zlib", "struct", "_md5",
    "cStringIO", "array", "binascii",
    # the following are needed for pyrepl (and hence for the
    # interactive prompt/pdb)
    "termios", "_minimal_curses",
])

reverse_debugger_disable_modules = set([
    "_continuation", "_vmprof", "_multiprocessing",
    "micronumpy",
    ])

# XXX this should move somewhere else, maybe to platform ("is this posixish"
#     check or something)
if sys.platform == "win32":
    working_modules.add("_winreg")
    # unix only modules
    for name in ["crypt", "fcntl", "pwd", "termios", "_minimal_curses"]:
        if name in working_modules:
            working_modules.remove(name)
        if name in translation_modules:
            translation_modules.remove(name)

    if "faulthandler" in working_modules:
        working_modules.remove("faulthandler")  # missing details
    if "_vmprof" in working_modules:
        working_modules.remove("_vmprof")  # FIXME: missing details

    # The _locale module is needed by site.py on Windows
    default_modules.add("_locale")

    # needed to invoke MSVC
    translation_modules.update(["thread", "_winreg", "_cffi_backend"])

    # not ported yet
    if IS_64_BITS:
        for name in ["cpyext", "_cppyy", "micronumpy"]:
            if name in working_modules:
                working_modules.remove(name)

if sys.platform == "sunos5":
    working_modules.remove('fcntl')  # LOCK_NB not defined
    working_modules.remove("_minimal_curses")
    working_modules.remove("termios")
    if "_cppyy" in working_modules:
        working_modules.remove("_cppyy")  # depends on ctypes

#if sys.platform.startswith("linux"):
#    _mach = os.popen('uname -m', 'r').read().strip()
#    if _mach.startswith(...):
#        working_modules.remove("_continuation")


module_dependencies = {
    '_multiprocessing': [('objspace.usemodules.time', True),
                         ('objspace.usemodules.thread', True)],
    'cpyext': [('objspace.usemodules.array', True)],
    '_cppyy': [('objspace.usemodules.cpyext', True)],
    'faulthandler': [('objspace.usemodules._vmprof', True)],
    }
module_suggests = {
    # the reason you want _rawffi is for ctypes, which
    # itself needs the interp-level struct module
    # because 'P' is missing from the app-level one
    "_rawffi": [("objspace.usemodules.struct", True)],
    "cpyext": [("translation.secondaryentrypoints", "cpyext,main")],
}
if sys.platform == "win32":
    module_suggests["cpyext"].append(("translation.shared", True))


pypy_optiondescription = OptionDescription("objspace", "Object Space Options", [
    OptionDescription("usemodules", "Which Modules should be used", [
        BoolOption(modname, "use module %s" % (modname, ),
                   default=modname in default_modules,
                   cmdline="--withmod-%s" % (modname, ),
                   requires=module_dependencies.get(modname, []),
                   suggests=module_suggests.get(modname, []),
                   negation=modname not in essential_modules,
                   )
        for modname in all_modules]),

    BoolOption("allworkingmodules", "use as many working modules as possible",
               # NB. defaults to True, but in py.py this is overridden by
               # a False suggestion because it takes a while to start up.
               # Actual module enabling only occurs if
               # enable_allworkingmodules() is called, and it depends
               # on the selected backend.
               default=True,
               cmdline="--allworkingmodules",
               negation=True),

    StrOption("extmodules",
              "Comma-separated list of third-party builtin modules",
              cmdline="--ext",
              default=None),

    BoolOption("translationmodules",
          "use only those modules that are needed to run translate.py on pypy",
               default=False,
               cmdline="--translationmodules",
               suggests=[("objspace.allworkingmodules", False)]),

    BoolOption("lonepycfiles", "Import pyc files with no matching py file",
               default=False),

    StrOption("soabi",
              "Tag to differentiate extension modules built for different Python interpreters",
              cmdline="--soabi",
              default=None),

    BoolOption("honor__builtins__",
               "Honor the __builtins__ key of a module dictionary",
               default=False),

    BoolOption("disable_call_speedhacks",
               "make sure that all calls go through space.call_args",
               default=False),

    BoolOption("disable_entrypoints",
               "Disable external entry points, notably the"
               " cpyext module and cffi's embedding mode.",
               default=False,
               requires=[("objspace.usemodules.cpyext", False)]),

    ChoiceOption("hash",
                 "The hash function to use for strings: fnv from CPython 2.7"
                 " or siphash24 from CPython >= 3.4",
                 ["fnv", "siphash24"],
                 default="fnv",
                 cmdline="--hash"),

    OptionDescription("std", "Standard Object Space Options", [
        BoolOption("withtproxy", "support transparent proxies",
                   default=True),

        BoolOption("withprebuiltint", "prebuild commonly used int objects",
                   default=False),

        IntOption("prebuiltintfrom", "lowest integer which is prebuilt",
                  default=-5, cmdline="--prebuiltintfrom"),

        IntOption("prebuiltintto", "highest integer which is prebuilt",
                  default=100, cmdline="--prebuiltintto"),

        BoolOption("withsmalllong", "use a version of 'long' in a C long long",
                   default=False),

        BoolOption("withspecialisedtuple",
                   "use specialised tuples",
                   default=False),

        BoolOption("withliststrategies",
                   "enable optimized ways to store lists of primitives ",
                   default=True),

        BoolOption("withmethodcachecounter",
                   "try to cache methods and provide a counter in __pypy__. "
                   "for testing purposes only.",
                   default=False),
        IntOption("methodcachesizeexp",
                  " 2 ** methodcachesizeexp is the size of the of the method cache ",
                  default=11),
        BoolOption("intshortcut",
                   "special case addition and subtraction of two integers in BINARY_ADD/"
                   "/BINARY_SUBTRACT and their inplace counterparts",
                   default=False),
        BoolOption("optimized_list_getitem",
                   "special case the 'list[integer]' expressions",
                   default=False),
        BoolOption("newshortcut",
                   "cache and shortcut calling __new__ from builtin types",
                   default=False),
        BoolOption("reinterpretasserts",
                   "Perform reinterpretation when an assert fails "
                   "(only relevant for tests)",
                   default=False),

     ]),
])

def get_pypy_config(overrides=None, translating=False):
    from rpython.config.translationoption import get_combined_translation_config
    return get_combined_translation_config(
            pypy_optiondescription, overrides=overrides,
            translating=translating)

def set_pypy_opt_level(config, level):
    """Apply PyPy-specific optimization suggestions on the 'config'.
    The optimizations depend on the selected level and possibly on the backend.
    """
    # all the good optimizations for PyPy should be listed here
    if level in ['2', '3', 'jit']:
        config.objspace.std.suggest(intshortcut=True)
        config.objspace.std.suggest(optimized_list_getitem=True)
        #config.objspace.std.suggest(newshortcut=True)
        config.objspace.std.suggest(withspecialisedtuple=True)
        #if not IS_64_BITS:
        #    config.objspace.std.suggest(withsmalllong=True)

    # extra costly optimizations only go in level 3
    if level == '3':
        config.translation.suggest(profopt=
            "-c 'from richards import main;main(); "
                "from test import pystone; pystone.main()'")

    # memory-saving optimizations
    if level == 'mem':
        config.objspace.std.suggest(withprebuiltint=True)
        config.objspace.std.suggest(withliststrategies=True)
        if not IS_64_BITS:
            config.objspace.std.suggest(withsmalllong=True)

    # extra optimizations with the JIT
    if level == 'jit':
        pass # none at the moment


def enable_allworkingmodules(config):
    modules = working_modules.copy()
    if config.translation.sandbox:
        modules = default_modules
    if config.translation.reverse_debugger:
        for mod in reverse_debugger_disable_modules:
            setattr(config.objspace.usemodules, mod, False)
    # ignore names from 'essential_modules', notably 'exceptions', which
    # may not be present in config.objspace.usemodules at all
    modules = [name for name in modules if name not in essential_modules]

    config.objspace.usemodules.suggest(**dict.fromkeys(modules, True))

def enable_translationmodules(config):
    modules = translation_modules
    modules = [name for name in modules if name not in essential_modules]
    config.objspace.usemodules.suggest(**dict.fromkeys(modules, True))


if __name__ == '__main__':
    config = get_pypy_config()
    print config.getpaths()
    parser = to_optparse(config) #, useoptions=["translation.*"])
    option, args = parser.parse_args()
    print config
    print working_modules



import random, os
try:
    import cffi
except ImportError:
    import py
    py.test.skip("cffi not installed")

ffi = cffi.FFI()

ffi.cdef("""
typedef struct {
    unsigned long key;
    char *data;
    ...;
} skipnode_t;

skipnode_t *skiplist_malloc(unsigned long datasize);
skipnode_t *skiplist_search(skipnode_t *head, unsigned long searchkey);
void skiplist_insert(skipnode_t *head, skipnode_t *new);
skipnode_t *skiplist_remove(skipnode_t *head, unsigned long exact_key);
unsigned long skiplist_firstkey(skipnode_t *head);
""")

filename = os.path.join(os.path.dirname(__file__), '..', 'src', 'skiplist.c')
lib = ffi.verify(open(filename).read())


def test_insert_search_remove():
    my_head = ffi.new("skipnode_t *")
    assert lib.skiplist_search(my_head, 0) == my_head
    #
    keys = random.sample(xrange(2, 10**9), 50000)
    nodes = {}
    for key in keys:
        node = lib.skiplist_malloc(4)
        node.key = key
        ffi.cast("int *", node.data)[0] = key
        lib.skiplist_insert(my_head, node)
        nodes[key] = node
    #
    random.shuffle(keys)
    for key in keys:
        node = lib.skiplist_search(my_head, key)
        assert nodes[key] == node
        if key + 1 not in nodes:
            assert node == lib.skiplist_search(my_head, key + 1)
    #
    keys.sort()
    following = {}
    preceeding = {}
    for key, next_key in zip(keys[:-1], keys[1:]):
        following[key] = next_key
        preceeding[next_key] = key
    following[0] = keys[0]
    following[keys[-1]] = 10**9
    preceeding[keys[0]] = 0
    #
    for i in range(100000):
        random_key = random.randrange(2, 10**9)
        node = lib.skiplist_search(my_head, random_key)
        assert node.key <= random_key
        if node == my_head:
            assert random_key < following[0]
        else:
            assert node == nodes[node.key]
            assert following[node.key] > random_key
    #
    random_keys = list(keys)
    random.shuffle(random_keys)
    for i in range(10000):
        node = nodes.pop(random_keys.pop())
        prev = preceeding[node.key]
        next = following[node.key]
        following[prev] = next
        preceeding[next] = prev
        res = lib.skiplist_remove(my_head, node.key)
        assert res == node
        if prev == 0:
            assert lib.skiplist_search(my_head, node.key) == my_head
        else:
            assert lib.skiplist_search(my_head, node.key) == nodes[prev]
        res = lib.skiplist_remove(my_head, node.key)
        assert res == ffi.NULL
    #
    for i in range(100000):
        random_key = random.randrange(2, 10**9)
        node = lib.skiplist_search(my_head, random_key)
        assert node.key <= random_key
        if node == my_head:
            assert random_key < following[0]
        else:
            assert node == nodes[node.key]
            assert following[node.key] > random_key



import pytest

import _functools


def test_partial_reduce():
    partial = _functools.partial(test_partial_reduce)
    state = partial.__reduce__()
    d = state[2][2]
    assert state == (type(partial), (test_partial_reduce,),
                     (test_partial_reduce, (), d, None))
    assert d is None or d == {}      # both are acceptable

def test_partial_setstate():
    partial = _functools.partial(object)
    partial.__setstate__((test_partial_setstate, (), None, None))
    assert partial.func == test_partial_setstate

def test_partial_pickle():
    import pickle
    partial1 = _functools.partial(test_partial_pickle)
    string = pickle.dumps(partial1)
    partial2 = pickle.loads(string)
    assert partial1.func == partial2.func

def test_immutable_attributes():
    partial = _functools.partial(object)
    with pytest.raises((TypeError, AttributeError)):
        partial.func = sum
    with pytest.raises(TypeError) as exc:
        del partial.__dict__
    assert str(exc.value) == "a partial object's dictionary may not be deleted"
    with pytest.raises(AttributeError):
        del partial.zzz

def test_self_keyword():
    partial = _functools.partial(dict, self=42)
    assert partial(other=43) == {'self': 42, 'other': 43}

def test_no_keywords():
    kw1 = _functools.partial(dict).keywords
    kw2 = _functools.partial(dict, **{}).keywords
    # CPython gives different results for these two cases, which is not
    # possible to emulate in pure Python; see issue #2043
    assert kw1 == {} or kw1 is None
    assert kw2 == {}



"""
Symbol tabling building.
"""

from pypy.interpreter.astcompiler import ast, misc
from pypy.interpreter.pyparser.error import SyntaxError

# These are for internal use only:
SYM_BLANK = 0
SYM_GLOBAL = 1
SYM_ASSIGNED = 2 # Or deleted actually.
SYM_PARAM = 2 << 1
SYM_USED = 2 << 2
SYM_BOUND = (SYM_PARAM | SYM_ASSIGNED)

# codegen.py actually deals with these:
SCOPE_UNKNOWN = 0
SCOPE_GLOBAL_IMPLICIT = 1
SCOPE_GLOBAL_EXPLICIT = 2
SCOPE_LOCAL = 3
SCOPE_FREE = 4
SCOPE_CELL = 5


class Scope(object):

    can_be_optimized = False

    def __init__(self, name, lineno=0, col_offset=0):
        self.lineno = lineno
        self.col_offset = col_offset
        self.parent = None
        self.name = name
        self.locals_fully_known = False
        self.symbols = None
        self.roles = {}
        self.varnames = []
        self.children = []
        self.free_vars = []    # a bag of names: the order doesn't matter here
        self.temp_name_counter = 1
        self.has_exec = False
        self.has_free = False
        self.child_has_free = False
        self.nested = False
        self.doc_removable = False
        self._in_try_body_depth = 0

    def error(self, msg, ast_node=None):
        if ast_node is None:
            lineno = self.lineno
            col_offset = self.col_offset
        else:
            lineno = ast_node.lineno
            col_offset = ast_node.col_offset
        raise SyntaxError(msg, lineno, col_offset)

    def lookup(self, name):
        """Find the scope of identifier 'name'."""
        return self.symbols.get(self.mangle(name), SCOPE_UNKNOWN)

    def lookup_role(self, name):
        return self.roles.get(self.mangle(name), SYM_BLANK)

    def new_temporary_name(self):
        """Return the next temporary name.

        This must be in sync with PythonCodeGenerator's counter.
        """
        self.note_symbol("_[%d]" % (self.temp_name_counter,), SYM_ASSIGNED)
        self.temp_name_counter += 1

    def note_symbol(self, identifier, role, ast_node=None):
        """Record that identifier occurs in this scope."""
        mangled = self.mangle(identifier)
        new_role = role
        if mangled in self.roles:
            old_role = self.roles[mangled]
            if old_role & SYM_PARAM and role & SYM_PARAM:
                err = "duplicate argument '%s' in function definition" % \
                    (identifier,)
                self.error(err, ast_node)
            new_role |= old_role
        self.roles[mangled] = new_role
        if role & SYM_PARAM:
            self.varnames.append(mangled)
        return mangled

    def note_try_start(self, try_node):
        """Called when a try is found, before visiting the body."""
        self._in_try_body_depth += 1

    def note_try_end(self, try_node):
        """Called after visiting a try body."""
        self._in_try_body_depth -= 1

    def note_yield(self, yield_node):
        """Called when a yield is found."""
        self.error("'yield' outside function", yield_node)

    def note_return(self, ret):
        """Called when a return statement is found."""
        self.error("return outside function", ret)

    def note_exec(self, exc):
        """Called when an exec statement is found."""
        self.has_exec = True

    def note_import_star(self, imp):
        """Called when a star import is found."""
        return False

    def mangle(self, name):
        if self.parent:
            return self.parent.mangle(name)
        else:
            return name

    def add_child(self, child_scope):
        """Note a new child scope."""
        child_scope.parent = self
        self.children.append(child_scope)

    def _finalize_name(self, name, flags, local, bound, free, globs):
        """Decide on the scope of a name."""
        if flags & SYM_GLOBAL:
            if flags & SYM_PARAM:
                err = "name '%s' is local and global" % (name,)
                self.error(err)
            self.symbols[name] = SCOPE_GLOBAL_EXPLICIT
            globs[name] = None
            if bound:
                try:
                    del bound[name]
                except KeyError:
                    pass
        elif flags & SYM_BOUND:
            self.symbols[name] = SCOPE_LOCAL
            local[name] = None
            try:
                del globs[name]
            except KeyError:
                pass
        elif bound and name in bound:
            self.symbols[name] = SCOPE_FREE
            self.free_vars.append(name)
            free[name] = None
            self.has_free = True
        elif name in globs:
            self.symbols[name] = SCOPE_GLOBAL_IMPLICIT
        else:
            if self.nested:
                self.has_free = True
            self.symbols[name] = SCOPE_GLOBAL_IMPLICIT

    def _pass_on_bindings(self, local, bound, globs, new_bound, new_globs):
        """Allow child scopes to see names bound here and in outer scopes."""
        new_globs.update(globs)
        if bound:
            new_bound.update(bound)

    def _finalize_cells(self, free):
        """Hook for FunctionScope."""
        pass

    def _check_optimization(self):
        pass

    _hide_bound_from_nested_scopes = False

    def finalize(self, bound, free, globs):
        """Enter final bookeeping data in to self.symbols."""
        self.symbols = {}
        local = {}
        new_globs = {}
        new_bound = {}
        new_free = {}
        if self._hide_bound_from_nested_scopes:
            self._pass_on_bindings(local, bound, globs, new_bound, new_globs)
        for name, flags in self.roles.iteritems():
            self._finalize_name(name, flags, local, bound, free, globs)
        if not self._hide_bound_from_nested_scopes:
            self._pass_on_bindings(local, bound, globs, new_bound, new_globs)
        child_frees = {}
        for child in self.children:
            # Symbol dictionaries are copied to avoid having child scopes
            # pollute each other's.
            child_free = new_free.copy()
            child.finalize(new_bound.copy(), child_free, new_globs.copy())
            child_frees.update(child_free)
            if child.has_free or child.child_has_free:
                self.child_has_free = True
        new_free.update(child_frees)
        self._finalize_cells(new_free)
        for name in new_free:
            try:
                role_here = self.roles[name]
            except KeyError:
                if name in bound:
                    self.symbols[name] = SCOPE_FREE
                    self.free_vars.append(name)
            else:
                if role_here & (SYM_BOUND | SYM_GLOBAL) and \
                        self._hide_bound_from_nested_scopes:
                    # This happens when a class level attribute or method has
                    # the same name as a free variable passing through the class
                    # scope.  We add the name to the class scope's list of free
                    # vars, so it will be passed through by the interpreter, but
                    # we leave the scope alone, so it can be local on its own.
                    self.free_vars.append(name)
        self._check_optimization()
        free.update(new_free)


class ModuleScope(Scope):

    def __init__(self):
        Scope.__init__(self, "top")


class FunctionScope(Scope):

    can_be_optimized = True

    def __init__(self, name, lineno, col_offset):
        Scope.__init__(self, name, lineno, col_offset)
        self.has_variable_arg = False
        self.has_keywords_arg = False
        self.is_generator = False
        self.has_yield_inside_try = False
        self.optimized = True
        self.return_with_value = False
        self.import_star = None
        self.bare_exec = None

    def note_yield(self, yield_node):
        if self.return_with_value:
            self.error("'return' with argument inside generator",
                       self.ret)
        self.is_generator = True
        if self._in_try_body_depth > 0:
            self.has_yield_inside_try = True

    def note_return(self, ret):
        if ret.value:
            if self.is_generator:
                self.error("'return' with argument inside generator", ret)
            self.return_with_value = True
            self.ret = ret

    def note_exec(self, exc):
        Scope.note_exec(self, exc)
        if not exc.globals:
            self.optimized = False
            self.bare_exec = exc

    def note_import_star(self, imp):
        self.optimized = False
        self.import_star = imp
        return True

    def note_variable_arg(self, vararg):
        self.has_variable_arg = True

    def note_keywords_arg(self, kwarg):
        self.has_keywords_arg = True

    def add_child(self, child_scope):
        Scope.add_child(self, child_scope)
        child_scope.nested = True

    def _pass_on_bindings(self, local, bound, globs, new_bound, new_globs):
        new_bound.update(local)
        Scope._pass_on_bindings(self, local, bound, globs, new_bound, new_globs)

    def _finalize_cells(self, free):
        for name, role in self.symbols.iteritems():
            if role == SCOPE_LOCAL and name in free:
                self.symbols[name] = SCOPE_CELL
                del free[name]

    def _check_optimization(self):
        if (self.has_free or self.child_has_free) and not self.optimized:
            err = None
            if self.child_has_free:
                trailer = "contains a nested function with free variables"
            else:
                trailer = "is a nested function"
            name = self.name
            if self.import_star:
                node = self.import_star
                if self.bare_exec:
                    err = "function '%s' uses import * and bare exec, " \
                        "which are illegal because it %s" % (name, trailer)
                else:
                    err = "import * is not allowed in function '%s' because " \
                        "it %s" % (name, trailer)
            elif self.bare_exec:
                node = self.bare_exec
                err = "unqualified exec is not allowed in function '%s' " \
                    "because it %s" % (name, trailer)
            else:
                raise AssertionError("unknown reason for unoptimization")
            self.error(err, node)
        self.locals_fully_known = self.optimized and not self.has_exec


class ClassScope(Scope):

    _hide_bound_from_nested_scopes = True

    def __init__(self, clsdef):
        Scope.__init__(self, clsdef.name, clsdef.lineno, clsdef.col_offset)

    def mangle(self, name):
        return misc.mangle(name, self.name)


class SymtableBuilder(ast.GenericASTVisitor):
    """Find symbol information from AST."""

    def __init__(self, space, module, compile_info):
        self.space = space
        self.module = module
        self.compile_info = compile_info
        self.scopes = {}
        self.scope = None
        self.stack = []
        top = ModuleScope()
        self.globs = top.roles
        self.push_scope(top, module)
        try:
            module.walkabout(self)
            top.finalize(None, {}, {})
        except SyntaxError as e:
            e.filename = compile_info.filename
            raise
        self.pop_scope()
        assert not self.stack

    def push_scope(self, scope, node):
        """Push a child scope."""
        if self.stack:
            self.stack[-1].add_child(scope)
        self.stack.append(scope)
        self.scopes[node] = scope
        # Convenience
        self.scope = scope

    def pop_scope(self):
        self.stack.pop()
        if self.stack:
            self.scope = self.stack[-1]
        else:
            self.scope = None

    def find_scope(self, scope_node):
        """Lookup the scope for a given AST node."""
        return self.scopes[scope_node]

    def implicit_arg(self, pos):
        """Note a implicit arg for implicit tuple unpacking."""
        name = ".%d" % (pos,)
        self.note_symbol(name, SYM_PARAM)

    def note_symbol(self, identifier, role, ast_node=None):
        """Note the identifer on the current scope."""
        mangled = self.scope.note_symbol(identifier, role, ast_node)
        if role & SYM_GLOBAL:
            if mangled in self.globs:
                role |= self.globs[mangled]
            self.globs[mangled] = role

    def visit_FunctionDef(self, func):
        self.note_symbol(func.name, SYM_ASSIGNED)
        # Function defaults and decorators happen in the outer scope.
        args = func.args
        assert isinstance(args, ast.arguments)
        self.visit_sequence(args.defaults)
        self.visit_sequence(func.decorator_list)
        new_scope = FunctionScope(func.name, func.lineno, func.col_offset)
        self.push_scope(new_scope, func)
        func.args.walkabout(self)
        self.visit_sequence(func.body)
        self.pop_scope()

    def visit_Return(self, ret):
        self.scope.note_return(ret)
        ast.GenericASTVisitor.visit_Return(self, ret)

    def visit_ClassDef(self, clsdef):
        self.note_symbol(clsdef.name, SYM_ASSIGNED)
        self.visit_sequence(clsdef.bases)
        self.visit_sequence(clsdef.decorator_list)
        self.push_scope(ClassScope(clsdef), clsdef)
        self.visit_sequence(clsdef.body)
        self.pop_scope()

    def visit_ImportFrom(self, imp):
        for alias in imp.names:
            if self._visit_alias(alias):
                if self.scope.note_import_star(imp):
                    msg = "import * only allowed at module level"
                    misc.syntax_warning(
                        self.space, msg, self.compile_info.filename,
                        imp.lineno, imp.col_offset)

    def _visit_alias(self, alias):
        assert isinstance(alias, ast.alias)
        if alias.asname:
            store_name = alias.asname
        else:
            store_name = alias.name
            if store_name == "*":
                return True
            dot = store_name.find(".")
            if dot > 0:
                store_name = store_name[:dot]
        self.note_symbol(store_name, SYM_ASSIGNED)
        return False

    def visit_alias(self, alias):
        self._visit_alias(alias)

    def visit_Exec(self, exc):
        self.scope.note_exec(exc)
        ast.GenericASTVisitor.visit_Exec(self, exc)

    def visit_Yield(self, yie):
        self.scope.note_yield(yie)
        ast.GenericASTVisitor.visit_Yield(self, yie)

    def visit_Global(self, glob):
        for name in glob.names:
            old_role = self.scope.lookup_role(name)
            if old_role & (SYM_USED | SYM_ASSIGNED):
                if old_role & SYM_ASSIGNED:
                    msg = "name '%s' is assigned to before global declaration" \
                        % (name,)
                else:
                    msg = "name '%s' is used prior to global declaration" % \
                        (name,)
                misc.syntax_warning(self.space, msg, self.compile_info.filename,
                                    glob.lineno, glob.col_offset)
            if old_role & SYM_PARAM:
                err = "name '%s' is local and global" % (name,)
                self.scope.error(err, glob)
            self.note_symbol(name, SYM_GLOBAL)

    def visit_Lambda(self, lamb):
        args = lamb.args
        assert isinstance(args, ast.arguments)
        self.visit_sequence(args.defaults)
        new_scope = FunctionScope("lambda", lamb.lineno, lamb.col_offset)
        self.push_scope(new_scope, lamb)
        lamb.args.walkabout(self)
        lamb.body.walkabout(self)
        self.pop_scope()

    def _visit_comprehension(self, node, comps, *consider):
        outer = comps[0]
        assert isinstance(outer, ast.comprehension)
        outer.iter.walkabout(self)
        new_scope = FunctionScope("genexp", node.lineno, node.col_offset)
        self.push_scope(new_scope, node)
        self.implicit_arg(0)
        outer.target.walkabout(self)
        self.visit_sequence(outer.ifs)
        self.visit_sequence(comps[1:])
        for item in list(consider):
            item.walkabout(self)
        self.pop_scope()

    def visit_GeneratorExp(self, genexp):
        self._visit_comprehension(genexp, genexp.generators, genexp.elt)

    def visit_SetComp(self, setcomp):
        self._visit_comprehension(setcomp, setcomp.generators, setcomp.elt)

    def visit_DictComp(self, dictcomp):
        self._visit_comprehension(dictcomp, dictcomp.generators,
                                  dictcomp.value, dictcomp.key)

    def visit_With(self, wih):
        self.scope.new_temporary_name()
        if wih.optional_vars:
            self.scope.new_temporary_name()
        wih.context_expr.walkabout(self)
        if wih.optional_vars:
            wih.optional_vars.walkabout(self)
        self.scope.note_try_start(wih)
        self.visit_sequence(wih.body)
        self.scope.note_try_end(wih)

    def visit_arguments(self, arguments):
        scope = self.scope
        assert isinstance(scope, FunctionScope) # Annotator hint.
        if arguments.args:
            self._handle_params(arguments.args, True)
        if arguments.vararg:
            self.note_symbol(arguments.vararg, SYM_PARAM)
            scope.note_variable_arg(arguments.vararg)
        if arguments.kwarg:
            self.note_symbol(arguments.kwarg, SYM_PARAM)
            scope.note_keywords_arg(arguments.kwarg)
        if arguments.args:
            self._handle_nested_params(arguments.args)

    def _handle_params(self, params, is_toplevel):
        for i in range(len(params)):
            arg = params[i]
            if isinstance(arg, ast.Name):
                self.note_symbol(arg.id, SYM_PARAM, arg)
            elif isinstance(arg, ast.Tuple):
                # Tuple unpacking in the argument list.  Add a secret variable
                # name to recieve the tuple with.
                if is_toplevel:
                    self.implicit_arg(i)
            else:
                raise AssertionError("unknown parameter type")
        if not is_toplevel:
            self._handle_nested_params(params)

    def _handle_nested_params(self, params):
        for param in params:
            if isinstance(param, ast.Tuple):
                self._handle_params(param.elts, False)

    def visit_Name(self, name):
        if name.ctx == ast.Load:
            role = SYM_USED
        else:
            role = SYM_ASSIGNED
        self.note_symbol(name.id, role)

    def visit_TryExcept(self, node):
        self.scope.note_try_start(node)
        self.visit_sequence(node.body)
        self.scope.note_try_end(node)
        self.visit_sequence(node.handlers)
        self.visit_sequence(node.orelse)

    def visit_TryFinally(self, node):
        self.scope.note_try_start(node)
        self.visit_sequence(node.body)
        self.scope.note_try_end(node)
        self.visit_sequence(node.finalbody)



import os
from rpython.tool.udir import udir
import tempfile
from rpython.jit.backend.arm.test.support import AS

class ASMInstruction(object):

    asm_opts = '-mfpu=neon -mcpu=cortex-a8 -march=armv7-a'
    body = """.section .text
.arm
_start: .global _start
        .global main
        b main
main:
    .ascii "START"
    %s
    .ascii "END"
"""
    begin_tag = 'START'
    end_tag = 'END'
    base_name = 'test_%d.asm' 
    index = 0

    def __init__(self, instr):
        self.instr = instr
        self.file = udir.join(self.base_name % self.index)
        while self.file.check():
            self.index += 1
            self.file = udir.join(self.base_name % self.index)

    def encode(self):
        f = open("%s/a.out" % (udir),'rb')
        data = f.read()
        #f.close()
        i = data.find(self.begin_tag)
        assert i>=0
        j = data.find(self.end_tag, i)
        assert j>=0
        as_code = data[i+len(self.begin_tag):j]
        return as_code



    def assemble(self, *args):
        res = self.body % (self.instr)
        self.file.write(res)
        os.system("%s --fatal-warnings %s %s -o %s/a.out" % (AS, self.asm_opts, self.file, udir))

    #def __del__(self):
    #    self.file.close()

def assemble(instr):
    a = ASMInstruction(instr)
    a.assemble(instr)
    return a.encode()



"""
This disables the backend tests on non riscV platforms.
Note that you need "--slow" to run translation tests.
"""
import os
import pytest
from rpython.jit.backend import detect_cpu

cpu = detect_cpu.autodetect()
IS_RISCV = cpu.startswith( 'riscv64')
THIS_DIR = os.path.dirname(__file__)

@pytest.hookimpl(tryfirst=True)
def pytest_ignore_collect(path, config):
    path = str(path)
    if not IS_RISCV:
        if os.path.commonprefix([path, THIS_DIR]) == THIS_DIR:  # workaround for bug in pytest<3.0.5
            return True

def pytest_collect_file():
    if not IS_RISCV:
        # We end up here when calling py.test .../test_foo.py with a wrong cpu
        # It's OK to kill the whole session with the following line
        pytest.skip("RISCV tests skipped: cpu is %r" % (cpu,))



from __future__ import print_function

from rpython.annotator.policy import AnnotatorPolicy
from rpython.jit.tl import tiny1
from rpython.jit.backend.hlinfo import highleveljitinfo


def entry_point(args):
    """Main entry point of the stand-alone executable:
    takes a list of strings and returns the exit code.
    """
    # store args[0] in a place where the JIT log can find it (used by
    # viewcode.py to know the executable whose symbols it should display)
    highleveljitinfo.sys_executable = args[0]
    if len(args) < 4:
        print("Usage: %s bytecode x y" % (args[0],))
        return 2
    bytecode = args[1]
    x = int(args[2])
    y = int(args[3])
    res = tiny1.ll_plus_minus(bytecode, x, y)
    print(res)
    return 0

def target(driver, args):
    return entry_point, None



import pytest

@pytest.fixture
def tempfile(tmpdir):
    return str(tmpdir / 'tempfile1')

def test_f_locals():
    import sys
    f = sys._getframe()
    assert f.f_locals is locals()

def test_f_globals():
    import sys
    f = sys._getframe()
    assert f.f_globals is globals()
    with pytest.raises(TypeError):
        f.f_globals = globals()

def test_f_builtins():
    import sys, __builtin__
    f = sys._getframe()
    assert f.f_builtins is __builtin__.__dict__

def test_f_code():
    def g():
        import sys
        f = sys._getframe()
        return f.f_code
    assert g() is g.func_code

def test_f_trace_del():
    import sys
    f = sys._getframe()
    del f.f_trace
    assert f.f_trace is None

def test_f_lineno():
    def g():
        import sys
        f = sys._getframe()
        x = f.f_lineno
        y = f.f_lineno
        z = f.f_lineno
        return [x, y, z]
    origin = g.func_code.co_firstlineno
    assert g() == [origin+3, origin+4, origin+5]

def test_f_lineno_set(tempfile):
    def tracer(f, *args):
        def y(f, *args):
            return y
        def x(f, *args):
            f.f_lineno += 1
            return y  # "return None" should have the same effect, but see
                        # test_local_trace_function_returning_None_ignored
        return x

    open    # force fetching of this name now

    def function():
        xyz
        with open(tempfile, 'w') as f:
            pass
        return 3

    import sys
    sys.settrace(tracer)
    function()
    sys.settrace(None)
    # assert did not crash

def test_f_lineno_set_firstline():
    seen = []
    def tracer(f, event, *args):
        seen.append((event, f.f_lineno))
        if len(seen) == 5:
            f.f_lineno = 1       # bug shown only when setting lineno to 1
        return tracer

    def g():
        import sys
        sys.settrace(tracer)
        exec "x=1\ny=x+1\nz=y+1\nt=z+1\ns=t+1\n" in {}
        sys.settrace(None)

    g()
    assert seen == [('call', 1),
                    ('line', 1),
                    ('line', 2),
                    ('line', 3),
                    ('line', 4),
                    ('line', 2),
                    ('line', 3),
                    ('line', 4),
                    ('line', 5),
                    ('return', 5)]

def test_f_back():
    import sys
    def f():
        assert sys._getframe().f_code.co_name == g()
    def g():
        return sys._getframe().f_back.f_code.co_name
    f()

def test_f_back_virtualref():
    import sys
    def f():
        return g()
    def g():
        return sys._getframe()
    frame = f()
    assert frame.f_back.f_code.co_name == 'f'

def test_f_exc_xxx():
    import sys

    class OuterException(Exception):
        pass
    class InnerException(Exception):
        pass

    def g(exc_info):
        f = sys._getframe()
        assert f.f_exc_type is None
        assert f.f_exc_value is None
        assert f.f_exc_traceback is None
        try:
            raise InnerException
        except:
            assert f.f_exc_type is exc_info[0]
            assert f.f_exc_value is exc_info[1]
            assert f.f_exc_traceback is exc_info[2]
    try:
        raise OuterException
    except:
        g(sys.exc_info())

def test_virtualref_through_traceback():
    import sys
    def g():
        try:
            raise ValueError
        except:
            _, _, tb = sys.exc_info()
        return tb
    def f():
        return g()
    #
    tb = f()
    assert tb.tb_frame.f_code.co_name == 'g'
    assert tb.tb_frame.f_back.f_code.co_name == 'f'

def test_trace_basic():
    import sys
    l = []
    class Tracer:
        def __init__(self, i):
            self.i = i
        def trace(self, frame, event, arg):
            l.append((self.i, frame.f_code.co_name, event, arg))
            if frame.f_code.co_name == 'g2':
                return None    # don't trace g2
            return Tracer(self.i+1).trace
    def g3(n):
        n -= 5
        return n
    def g2(n):
        n += g3(2)
        n += g3(7)
        return n
    def g(n):
        n += g2(3)
        return n
    def f(n):
        n = g(n)
        return n * 7
    sys.settrace(Tracer(0).trace)
    x = f(4)
    sys.settrace(None)
    assert x == 42
    print l
    assert l == [(0, 'f', 'call', None),
                    (1, 'f', 'line', None),
                        (0, 'g', 'call', None),
                        (1, 'g', 'line', None),
                            (0, 'g2', 'call', None),
                                (0, 'g3', 'call', None),
                                (1, 'g3', 'line', None),
                                (2, 'g3', 'line', None),
                                (3, 'g3', 'return', -3),
                                (0, 'g3', 'call', None),
                                (1, 'g3', 'line', None),
                                (2, 'g3', 'line', None),
                                (3, 'g3', 'return', 2),
                        (2, 'g', 'line', None),
                        (3, 'g', 'return', 6),
                    (2, 'f', 'line', None),
                    (3, 'f', 'return', 42)]

def test_trace_exc():
    import sys
    l = []
    def ltrace(a,b,c):
        if b == 'exception':
            l.append(c)
        return ltrace
    def trace(a,b,c): return ltrace
    def f():
        try:
            raise Exception
        except:
            pass
    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(l) == 1
    assert isinstance(l[0][1], Exception)

def test_trace_ignore_hidden():
    import sys
    import _testing

    l = []
    def trace(a,b,c):
        l.append((a,b,c))

    def f():
        h = _testing.Hidden()
        r = h.meth()
        return r

    sys.settrace(trace)
    res = f()
    sys.settrace(None)
    assert len(l) == 1
    assert l[0][1] == 'call'
    assert res == 'hidden' # sanity

def test_trace_hidden_prints(tempfile):
    import sys

    l = []
    def trace(a,b,c):
        l.append((a,b,c))
        return trace

    outputf = open(tempfile, 'w')
    def f():
        print >> outputf, 1
        print >> outputf, 2
        print >> outputf, 3
        return "that's the return value"

    sys.settrace(trace)
    f()
    sys.settrace(None)
    outputf.close()
    # should get 1 "call", 3 "line" and 1 "return" events, and no call
    # or return for the internal app-level implementation of 'print'
    assert len(l) == 6
    assert [what for (frame, what, arg) in l] == [
        'call', 'line', 'line', 'line', 'line', 'return']
    assert l[-1][2] == "that's the return value"

def test_trace_return_exc():
    import sys
    l = []
    def trace(a,b,c):
        if b in ('exception', 'return'):
            l.append((b, c))
        return trace

    def g():
        raise Exception
    def f():
        try:
            g()
        except:
            pass
    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(l) == 4
    assert l[0][0] == 'exception'
    assert isinstance(l[0][1][1], Exception)
    assert l[1] == ('return', None)
    assert l[2][0] == 'exception'
    assert isinstance(l[2][1][1], Exception)
    assert l[3] == ('return', None)

def test_trace_raises_on_return():
    import sys
    def trace(frame, event, arg):
        if event == 'return':
            raise ValueError
        else:
            return trace

    def f(): return 1

    for i in xrange(sys.getrecursionlimit() + 1):
        sys.settrace(trace)
        try:
            f()
        except ValueError:
            pass

def test_trace_try_finally():
    import sys
    l = []
    def trace(frame, event, arg):
        if event == 'exception':
            l.append(arg)
        return trace

    def g():
        try:
            raise Exception
        finally:
            pass

    def f():
        try:
            g()
        except:
            pass

    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(l) == 2
    assert issubclass(l[0][0], Exception)
    assert issubclass(l[1][0], Exception)

def test_trace_raise_three_arg():
    import sys
    l = []
    def trace(frame, event, arg):
        if event == 'exception':
            l.append(arg)
        return trace

    def g():
        try:
            raise Exception
        except Exception as e:
            import sys
            raise Exception, e, sys.exc_info()[2]

    def f():
        try:
            g()
        except:
            pass

    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(l) == 2
    assert issubclass(l[0][0], Exception)
    assert issubclass(l[1][0], Exception)

def test_trace_generator_finalisation():
    import sys
    l = []
    got_exc = []
    def trace(frame, event, arg):
        l.append((frame.f_lineno, event))
        if event == 'exception':
            got_exc.append(arg)
        return trace

    d = {}
    exec """if 1:
    def g():
        try:
            yield True
        finally:
            pass

    def f():
        try:
            gen = g()
            gen.next()
            gen.close()
        except:
            pass
    """ in d
    f = d['f']

    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(got_exc) == 1
    assert issubclass(got_exc[0][0], GeneratorExit)
    assert l == [(8, 'call'),
                    (9, 'line'),
                    (10, 'line'),
                    (11, 'line'),
                    (2, 'call'),
                    (3, 'line'),
                    (4, 'line'),
                    (4, 'return'),
                    (12, 'line'),
                    (4, 'call'),
                    (4, 'exception'),
                    (6, 'line'),
                    (6, 'return'),
                    (12, 'return')]

def test_dont_trace_on_reraise():
    import sys
    l = []
    def ltrace(a,b,c):
        if b == 'exception':
            l.append(c)
        return ltrace
    def trace(a,b,c): return ltrace
    def f():
        try:
            1/0
        except:
            try:
                raise
            except:
                pass
    sys.settrace(trace)
    f()
    sys.settrace(None)
    assert len(l) == 1
    assert issubclass(l[0][0], Exception)

def test_dont_trace_on_raise_with_tb():
    import sys
    l = []
    def ltrace(a,b,c):
        if b == 'exception':
            l.append(c)
        return ltrace
    def trace(a,b,c): return ltrace
    def f():
        try:
            raise Exception
        except:
            return sys.exc_info()
    def g():
        exc, val, tb = f()
        try:
            raise exc, val, tb
        except:
            pass
    sys.settrace(trace)
    g()
    sys.settrace(None)
    assert len(l) == 1
    assert isinstance(l[0][1], Exception)

def test_trace_changes_locals():
    import sys
    def trace(frame, what, arg):
        frame.f_locals['x'] = 42
        return trace
    def f(x):
        return x
    sys.settrace(trace)
    res = f(1)
    sys.settrace(None)
    assert res == 42

def test_set_unset_f_trace():
    import sys
    seen = []
    def trace1(frame, what, arg):
        seen.append((1, frame, frame.f_lineno, what, arg))
        return trace1
    def trace2(frame, what, arg):
        seen.append((2, frame, frame.f_lineno, what, arg))
        return trace2
    def set_the_trace(f):
        f.f_trace = trace1
        sys.settrace(trace2)
        len(seen)     # take one line: should not be traced
    f = sys._getframe()
    set_the_trace(f)
    len(seen)     # take one line: should not be traced
    len(seen)     # take one line: should not be traced
    sys.settrace(None)   # and this line should be the last line traced
    len(seen)     # take one line
    del f.f_trace
    len(seen)     # take one line
    firstline = set_the_trace.func_code.co_firstlineno
    assert seen == [(1, f, firstline + 6, 'line', None),
                    (1, f, firstline + 7, 'line', None),
                    (1, f, firstline + 8, 'line', None)]

def test_locals2fast_freevar_bug():
    import sys
    def f(n):
        class A(object):
            def g(self):
                return n
            n = 42
        return A()
    res = f(10).g()
    assert res == 10
    #
    def trace(*args):
        return trace
    sys.settrace(trace)
    res = f(10).g()
    sys.settrace(None)
    assert res == 10

def test_throw_trace_bug():
    import sys
    def f():
        yield 5
    gen = f()
    assert next(gen) == 5
    seen = []
    def trace_func(frame, event, *args):
        seen.append(event)
        return trace_func
    sys.settrace(trace_func)
    try:
        gen.throw(ValueError)
    except ValueError:
        pass
    sys.settrace(None)
    assert seen == ['call', 'exception', 'return']

def test_generator_trace_stopiteration():
    import sys
    def f():
        yield 5
    gen = f()
    assert next(gen) == 5
    seen = []
    def trace_func(frame, event, *args):
        print('TRACE:', frame, event, args)
        seen.append(event)
        return trace_func
    def g():
        for x in gen:
            never_entered
    sys.settrace(trace_func)
    g()
    sys.settrace(None)
    print 'seen:', seen
    # on Python 3 we get an extra 'exception' when 'for' catches
    # StopIteration
    assert seen == ['call', 'line', 'call', 'return', 'return']

def test_local_trace_function_returning_None_ignored():
    # behave the same as CPython does, and in contradiction with
    # the documentation.
    def tracer(f, event, arg):
        assert event == 'call'
        return local_tracer

    seen = []
    def local_tracer(f, event, arg):
        seen.append(event)
        return None     # but 'local_tracer' will be called again

    def function():
        a = 1
        a = 2
        a = 3

    import sys
    sys.settrace(tracer)
    function()
    sys.settrace(None)
    assert seen == ["line", "line", "line", "return"]



"""
Helpers to pack and unpack a unicode character into raw bytes.
"""

import sys

UNICODE_SIZE = 4
BIGENDIAN = sys.byteorder == "big"

def pack_unichar(unich, buf, pos):
    pack_codepoint(ord(unich), buf, pos)

def pack_codepoint(unich, buf, pos):
    if UNICODE_SIZE == 2:
        if BIGENDIAN:
            buf.setitem(pos,   chr(unich >> 8))
            buf.setitem(pos+1, chr(unich & 0xFF))
        else:
            buf.setitem(pos,   chr(unich & 0xFF))
            buf.setitem(pos+1, chr(unich >> 8))
    else:
        if BIGENDIAN:
            buf.setitem(pos,   chr(unich >> 24))
            buf.setitem(pos+1, chr((unich >> 16) & 0xFF))
            buf.setitem(pos+2, chr((unich >> 8) & 0xFF))
            buf.setitem(pos+3, chr(unich & 0xFF))
        else:
            buf.setitem(pos,   chr(unich & 0xFF))
            buf.setitem(pos+1, chr((unich >> 8) & 0xFF))
            buf.setitem(pos+2, chr((unich >> 16) & 0xFF))
            buf.setitem(pos+3, chr(unich >> 24))

def unpack_codepoint(rawstring):
    assert len(rawstring) == UNICODE_SIZE
    if UNICODE_SIZE == 2:
        if BIGENDIAN:
            n = (ord(rawstring[0]) << 8 |
                 ord(rawstring[1]))
        else:
            n = (ord(rawstring[0]) |
                 ord(rawstring[1]) << 8)
    else:
        if BIGENDIAN:
            n = (ord(rawstring[0]) << 24 |
                 ord(rawstring[1]) << 16 |
                 ord(rawstring[2]) << 8 |
                 ord(rawstring[3]))
        else:
            n = (ord(rawstring[0]) |
                 ord(rawstring[1]) << 8 |
                 ord(rawstring[2]) << 16 |
                 ord(rawstring[3]) << 24)
    return n

def unpack_unichar(rawstring):
    return unichr(unpack_codepoint(rawstring))



import errno
from pypy.interpreter.error import oefmt
from pypy.module.cpyext.api import cpython_api, INTP_real
from pypy.module.cpyext.pyobject import PyObject
from rpython.rlib import rdtoa
from rpython.rlib import rfloat
from rpython.rlib import rposix, jit
from rpython.rlib.rarithmetic import intmask
from rpython.rtyper.lltypesystem import lltype
from rpython.rtyper.lltypesystem import rffi


# PyOS_double_to_string's "type", if non-NULL, will be set to one of:
Py_DTST_FINITE = 0
Py_DTST_INFINITE = 1
Py_DTST_NAN = 2

# Match the "type" back to values in CPython
DOUBLE_TO_STRING_TYPES_MAP = {
    rfloat.DIST_FINITE: Py_DTST_FINITE,
    rfloat.DIST_INFINITY: Py_DTST_INFINITE,
    rfloat.DIST_NAN: Py_DTST_NAN
}

@cpython_api([rffi.CONST_CCHARP, rffi.CCHARPP, PyObject], rffi.DOUBLE, error=-1.0)
@jit.dont_look_inside       # direct use of _get_errno()
def PyOS_string_to_double(space, s, endptr, w_overflow_exception):
    """Convert a string s to a double, raising a Python
    exception on failure.  The set of accepted strings corresponds to
    the set of strings accepted by Python's float() constructor,
    except that s must not have leading or trailing whitespace.
    The conversion is independent of the current locale.

    If endptr is NULL, convert the whole string.  Raise
    ValueError and return -1.0 if the string is not a valid
    representation of a floating-point number.

    If endptr is not NULL, convert as much of the string as
    possible and set *endptr to point to the first unconverted
    character.  If no initial segment of the string is the valid
    representation of a floating-point number, set *endptr to point
    to the beginning of the string, raise ValueError, and return
    -1.0.

    If s represents a value that is too large to store in a float
    (for example, "1e500" is such a string on many platforms) then
    if overflow_exception is NULL return Py_HUGE_VAL (with
    an appropriate sign) and don't set any exception.  Otherwise,
    overflow_exception must point to a Python exception object;
    raise that exception and return -1.0.  In both cases, set
    *endptr to point to the first character after the converted value.

    If any other error occurs during the conversion (for example an
    out-of-memory error), set the appropriate Python exception and
    return -1.0.
    """
    user_endptr = True
    try:
        if not endptr:
            endptr = lltype.malloc(rffi.CCHARPP.TO, 1, flavor='raw')
            user_endptr = False
        result = rdtoa.dg_strtod(s, endptr)
        endpos = (rffi.cast(rffi.LONG, endptr[0]) -
                  rffi.cast(rffi.LONG, s))
        if endpos == 0 or (not user_endptr and not endptr[0][0] == '\0'):
            low = rffi.constcharp2str(s).lower()
            sz = 0
            if len(low) < 3:
                pass
            elif low[0] == '-':
                if low.startswith('-infinity'):
                    result = -rfloat.INFINITY
                    sz = len("-infinity")
                elif low.startswith("-inf"):
                    result = -rfloat.INFINITY
                    sz = 4
                elif low.startswith("-nan"):
                    result = -rfloat.NAN
                    sz = 4
            elif low[0] == '+':
                if low.startswith("+infinity"):
                    result = rfloat.INFINITY
                    sz = len("+infinity")
                elif low.startswith("+inf"):
                    result = rfloat.INFINITY
                    sz = 4
                elif low.startswith("+nan"):
                    result = rfloat.NAN
                    sz = 4
            elif low.startswith("infinity"):
                result = rfloat.INFINITY
                sz = len("infinity")
            elif low.startswith("inf"):
                result = rfloat.INFINITY
                sz = 3
            elif low.startswith("nan"):
                result = rfloat.NAN
                sz = 3
            # result is set to 0.0 for a parse_error in dtoa.c
            # if it changed, we must have sucessfully converted
            if result != 0.0:
                if endptr:
                    endptr[0] = rffi.cast(rffi.CCHARP, rffi.ptradd(s, sz))
                return result
            raise oefmt(space.w_ValueError,
                        "invalid input at position %d", endpos)
        err = rffi.cast(lltype.Signed, rposix._get_errno())
        if err == errno.ERANGE:
            rposix._set_errno(rffi.cast(rffi.INT, 0))
            if w_overflow_exception is None:
                if result > 0:
                    return rfloat.INFINITY
                else:
                    return -rfloat.INFINITY
            else:
                raise oefmt(w_overflow_exception, "value too large")
        return result
    finally:
        if not user_endptr:
            lltype.free(endptr, flavor='raw')

@cpython_api([rffi.DOUBLE, lltype.Char, rffi.INT_real, rffi.INT_real, INTP_real], rffi.CCHARP)
def PyOS_double_to_string(space, val, format_code, precision, flags, ptype):
    """Convert a double val to a string using supplied
    format_code, precision, and flags.

    format_code must be one of 'e', 'E', 'f', 'F',
    'g', 'G' or 'r'.  For 'r', the supplied precision
    must be 0 and is ignored.  The 'r' format code specifies the
    standard repr() format.

    flags can be zero or more of the values Py_DTSF_SIGN,
    Py_DTSF_ADD_DOT_0, or Py_DTSF_ALT, or-ed together:

    Py_DTSF_SIGN means to always precede the returned string with a sign
    character, even if val is non-negative.

    Py_DTSF_ADD_DOT_0 means to ensure that the returned string will not look
    like an integer.

    Py_DTSF_ALT means to apply "alternate" formatting rules.  See the
    documentation for the PyOS_snprintf() '#' specifier for
    details.

    If ptype is non-NULL, then the value it points to will be set to one of
    Py_DTST_FINITE, Py_DTST_INFINITE, or Py_DTST_NAN, signifying that
    val is a finite number, an infinite number, or not a number, respectively.

    The return value is a pointer to buffer with the converted string or
    NULL if the conversion failed. The caller is responsible for freeing the
    returned string by calling PyMem_Free().
    """
    buffer, rtype = rfloat.double_to_string(val, format_code,
                                            intmask(precision),
                                            intmask(flags))
    if ptype != lltype.nullptr(INTP_real.TO):
        ptype[0] = rffi.cast(rffi.INT_real, DOUBLE_TO_STRING_TYPES_MAP[rtype])
    bufp = rffi.str2charp(buffer)
    return bufp
